# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

# INFORMATION ABOUT THE SEMPER FORMAT LABEL:
# Picture labels consist of a sequence of bytes
# ---------------------------------------------
# v61:256B | v7:at least 256B, rounded up to multiple of block size 64,
#               with max set by LNLAB in params.f
# The versions have the same contents for the first 256B, as set out below,
# and referred to as the label 'base'; beyond this, in the label 'extension',
# the structure is deliberately undefined, allowing you to make your own use
# of the additional storage.
#
# From Aug14:
#  B1-6    S e m p e r (ic chars)
#   7,8    ncol msb,lsb (BIG-ended)
#   9,10   nrow msb,lsb
#  11,12   nlay msb,lsb
#  13,14   ccol msb,lsb
#  15,16   crow msb,lsb
#  17,18   clay msb,lsb
#    19    class: 1-20
#            for image,macro,fou,spec,correln,undef,walsh,histog,plist,lut
#    20    form: 0,1,2,3,4 = byte,i*2,fp,com,i*4 from Aug08
#    21    wp: non-zero if prot
#  22-27   creation year(-1900?),month,day,hour,min,sec
# 28    v61|v7 # chars in range record | 255
#  29-55   v61: min,max values present (ic chars for decimal repn)
#           v7: min,max vals as two Fp values in B29-36 (LE ordered)
#               followed by 19 unused bytes B37-55
#    56    plist type: 1,2,3 = list,opencurve,closedcurve
#            acc to EXAMPD - code appears to use explicit numbers
# 57,58,59 ncol,nrow,nlay hsb
# 60,61,62 ccol,crow,clay hsb
#    63    RealCoords flag (non-zero -> DX,DY,DZ,X0,Y0,Z0,units held as below)
# 64    v61:0 | v7: # blocks in (this) picture label (incl extn)
#  65-67   0 (free at present)
#  68-71   DATA cmd V7   4-byte Fp values, order LITTLE-ended
#  72-75   DATA cmd V6
#  76-79   RealCoord DZ / V5  RealCoord pars as 4-byte Fp values, LITTLE-ended
#  80-83   ...       Z0 / V4
#  84-87   ...       DY / V3
#  88-91   ...       Y0 / V2
#  92-95   ...       DX / V1
#  96-99   ...       X0 / V0
# 100    # chars in title
# 101-244  title (ic chars)
# 245-248  RealCoord X unit (4 ic chars)
# 249-252  RealCoord Y unit (4 ic chars)
# 253-256  RealCoord Z unit (4 ic chars)
#
# Aug08-Aug14 labels held no RealCoord information, so:
#   B63    'free' and zero - flagging absence of RealCoord information
# 101-256  title (12 chars longer than now)
#
# Before Aug08 labels held no hsb for pic sizes, so:
#  57-99   all free/zero except for use by DATA cmd
# 101-256  title (ic chars)

from collections import OrderedDict
import struct
from functools import partial
import logging
import warnings
from datetime import datetime

import numpy as np
from traits.api import Undefined

from hyperspy.misc.array_tools import sarray2dict


_logger = logging.getLogger(__name__)


# Plugin characteristics
# ----------------------
format_name = 'SEMPER UNF (unformatted)'
description = 'Read data from SEMPER UNF files.'
full_support = True  # Hopefully?
# Recognised file extension
file_extensions = ('unf', 'UNF')
default_extension = 0
# Writing features
writes = [(1, 0), (1, 1), (1, 2), (2, 0), (2, 1)]  # All up to 3D
# ----------------------


class SemperFormat(object):

    """Class for importing and exporting SEMPER `.unf`-files.

    The :class:`~.SemperFormat` class represents a SEMPER binary file format
    with a header, which holds additional information. `.unf`-files can be
    saved and read from files.

    Attributes
    ----------
    data : :class:`~numpy.ndarray` (N=3)
        The phase map or magnetization information in a 3D array (with one
        slice).
    title : string
        Title of the file (not to be confused with the filename).
    offsets : tuple (N=3) of floats
        Offset shifts (in nm) of the grid origin (does not have to start at 0)
        in x, y, z.
    scales : tuple (N=3) of floats
        Grid spacing (nm per pixel) in x, y, z.
    units : tuple (N=3) of strings
        Units of the grid in x, y, z.
    metadata : dictionary
        A dictionary of all flags and metadata present in the `.unf`-file.

    """

    ICLASS_DICT = {1: 'image', 2: 'macro', 3: 'fourier', 4: 'spectrum',
                   5: 'correlation', 6: Undefined, 7: 'walsh',
                   8: 'position list', 9: 'histogram',
                   10: 'display look-up table'}

    ICLASS_DICT_INV = {v: k for k, v in ICLASS_DICT.items()}

    IFORM_DICT = {
        0: np.byte,
        1: np.int16,
        2: np.float32,
        3: np.complex64,
        4: np.int32}

    IFORM_DICT_INV = {v: k for k, v in IFORM_DICT.items()}

    HEADER_DTYPES = [('NCOL', '<i2'),
                     ('NROW', '<i2'),
                     ('NLAY', '<i2'),
                     ('ICLASS', '<i2'),
                     ('IFORM', '<i2'),
                     ('IFLAG', '<i2'),
                     ('IFORM', '<i2')]

    LABEL_DTYPES = [('SEMPER', ('<i2', 6)),   # Bytes 1-6
                    ('NCOL', ('<i2', 2)),     # Bytes 7,8
                    ('NROW', ('<i2', 2)),     # Bytes 9,10
                    ('NLAY', ('<i2', 2)),     # Bytes 11,12
                    ('ICCOLN', ('<i2', 2)),   # Bytes 13,14
                    ('ICROWN', ('<i2', 2)),   # Bytes 15,16
                    ('ICLAYN', ('<i2', 2)),   # Bytes 17,18
                    ('ICLASS', '<i2'),        # Bytes 19
                    ('IFORM', '<i2'),         # Bytes 20
                    ('IWP', '<i2'),           # Bytes 21
                    ('DATE', ('<i2', 6)),     # Bytes 22-27
                    ('NCRANG', '<i2'),        # Bytes 28
                    ('RANGE', ('<i2', 27)),   # Bytes 29-55
                    ('IPLTYP', '<i2'),        # Bytes 56
                    ('NCOLH', '<i2'),         # Bytes 57
                    ('NROWH', '<i2'),         # Bytes 58
                    ('NLAYH', '<i2'),         # Bytes 59
                    ('ICCOLNH', '<i2'),       # Bytes 60
                    ('ICROWNH', '<i2'),       # Bytes 61
                    ('ICLAYNH', '<i2'),       # Bytes 62
                    ('REALCO', '<i2'),        # Bytes 63
                    ('NBLOCK', '<i2'),        # Bytes 64
                    ('FREE', ('<i2', 3)),     # Bytes 65-67
                    ('DATAV6', ('<i2', 4)),   # Bytes 68-71
                    ('DATAV7', ('<i2', 4)),   # Bytes 72-75
                    ('DZV5', ('<i2', 4)),     # Bytes 76-79
                    ('Z0V4', ('<i2', 4)),     # Bytes 80-83
                    ('DYV3', ('<i2', 4)),     # Bytes 84-87
                    ('Y0V2', ('<i2', 4)),     # Bytes 88-91
                    ('DXV1', ('<i2', 4)),     # Bytes 92-95
                    ('X0V0', ('<i2', 4)),     # Bytes 96-99
                    ('NTITLE', '<i2'),        # Bytes 100
                    ('TITLE', ('<i2', 144)),  # Bytes 101-244
                    ('XUNIT', ('<i2', 4)),    # Bytes 245-248
                    ('YUNIT', ('<i2', 4)),    # Bytes 249-252
                    ('ZUNIT', ('<i2', 4))]    # Bytes 253-256

    def __init__(self,
                 data,
                 title='',
                 offsets=(0., 0., 0.),
                 scales=(1., 1., 1.),
                 units=(Undefined, Undefined, Undefined),
                 metadata=None):
        if metadata is None:
            metadata = {}
        # Make sure data is 3D!
        data = data[tuple(None for _ in range(3 - len(data.shape)))]
        self.data = data
        self.title = title
        self.offsets = offsets
        self.scales = scales
        self.units = units
        self.metadata = metadata
        _logger.debug('Created ' + str(self))

    @classmethod
    def _read_label(cls, unf_file):
        unpack = partial(
            unpack_from_intbytes,
            '<f')  # Unpacking function for 4 byte floats!
        rec_length = np.fromfile(
            unf_file,
            dtype='<i',
            count=1)[0]  # length of label
        label = sarray2dict(
            np.fromfile(
                unf_file,
                dtype=cls.LABEL_DTYPES,
                count=1))
        label['SEMPER'] = ''.join([str(chr(l)) for l in label['SEMPER']])
        assert label['SEMPER'] == 'Semper'
        # Process dimensions:
        for key in ['NCOL', 'NROW', 'NLAY', 'ICCOLN', 'ICROWN', 'ICLAYN']:
            value = 256**2 * \
                label.pop(key + 'H') + 256 * label[key][0] + label[key][1]
            label[key] = value
        # Process date:
        date = '{}-{}-{} {}:{}:{}'.format(label['DATE'][0] +
                                          1900, *
                                          label['DATE'][1:])
        label['DATE'] = date
        # Process range:
        if label['NCRANG'] == 255:
            range_min = unpack(label['RANGE'][:4])
            range_max = unpack(label['RANGE'][4:8])
            range_string = '{:.6g},{:.6g}'.format(range_min, range_max)
        else:
            range_string = ''.join([str(chr(l))
                                    for l in label['RANGE'][:label['NCRANG']]])
        label['RANGE'] = range_string
        # Process real coords:
        x0 = unpack(label.pop('X0V0'))
        dx = unpack(label.pop('DXV1'))
        y0 = unpack(label.pop('Y0V2'))
        dy = unpack(label.pop('DYV3'))
        z0 = unpack(label.pop('Z0V4'))
        dz = unpack(label.pop('DZV5'))
        if label['REALCO'] == 1:
            label.update({'X0V0': x0, 'Y0V2': y0, 'Z0V4': z0,
                          'DXV1': dx, 'DYV3': dy, 'DZV5': dz})
        # Process additional commands (unused, not sure about the purpose):
        data_v6 = unpack(label['DATAV6'])
        data_v7 = unpack(label['DATAV7'])
        label['DATAV6'] = data_v6
        label['DATAV7'] = data_v7
        # Process title:
        title = ''.join([str(chr(l))
                         for l in label['TITLE'][:label['NTITLE']]])
        label['TITLE'] = title
        # Process units:
        label['XUNIT'] = ''.join(
            [chr(l) for l in label['XUNIT']]).replace('\x00', '')
        label['YUNIT'] = ''.join(
            [chr(l) for l in label['YUNIT']]).replace('\x00', '')
        label['ZUNIT'] = ''.join(
            [chr(l) for l in label['ZUNIT']]).replace('\x00', '')
        # Sanity check:
        assert np.fromfile(unf_file, dtype='<i4', count=1)[0] == rec_length
        # Return label:
        return label

    def _get_label(self):
        pack = partial(
            pack_to_intbytes,
            '<f')  # Packing function for 4 byte floats!
        nlay, nrow, ncol = self.data.shape
        data, iform = self._check_format(self.data)
        title = self.title
        # Create label:
        label = np.zeros((1,), dtype=self.LABEL_DTYPES)
        # Fill label:
        label['SEMPER'] = [ord(c) for c in 'Semper']
        label['NCOLH'], remain = divmod(ncol, 256**2)
        label['NCOL'] = divmod(remain, 256)
        label['NROWH'], remain = divmod(nrow, 256**2)
        label['NROW'] = divmod(remain, 256)
        label['NLAYH'], remain = divmod(nlay, 256**2)
        label['NLAY'] = divmod(remain, 256)
        iccoln = self.metadata.get('ICCOLN', self.data.shape[2] // 2 + 1)
        label['ICCOLNH'], remain = divmod(iccoln, 256**2)
        label['ICCOLN'] = divmod(remain, 256)
        icrown = self.metadata.get('ICROWN', self.data.shape[1] // 2 + 1)
        label['ICROWNH'], remain = divmod(icrown, 256**2)
        label['ICROWN'] = divmod(remain, 256)
        iclayn = self.metadata.get('ICLAYN', self.data.shape[0] // 2 + 1)
        label['ICLAYNH'], remain = divmod(iclayn, 256**2)
        label['ICLAYN'] = divmod(remain, 256)
        label['ICLASS'] = self.metadata.get('ICLASS', 6)  # 6: Undefined!
        label['IFORM'] = iform
        label['IWP'] = self.metadata.get('IWP', 0)  # seems standard
        date = self.metadata.get('DATE', "%s" % datetime.now())
        year, time = date.split(' ')
        date_ints = (list(map(int, year.split('-'))) +
                     list(map(int, time.split(':'))))
        date_ints[0] -= 1900  # Modify year integer!
        label['DATE'] = date_ints
        range_string = '{:.4g},{:.4g}'.format(self.data.min(), self.data.max())
        label['NCRANG'] = len(range_string)
        label['RANGE'][:, :len(range_string)] = [ord(c) for c in range_string]
        label['IPLTYP'] = self.metadata.get('IPLTYP', 248)  # seems standard
        label['REALCO'] = 1  # Real coordinates are used!
        label['NBLOCK'] = 4  # Always 4 64b blocks!
        label['FREE'] = [0, 0, 0]
        label['DATAV6'] = pack(0.)  # Not used!
        label['DATAV7'] = pack(0.)  # Not used!
        label['DZV5'] = pack(self.scales[2])   # DZ
        label['Z0V4'] = pack(self.offsets[2])  # Z0
        label['DYV3'] = pack(self.scales[1])   # DY
        label['Y0V2'] = pack(self.offsets[1])  # Y0
        label['DXV1'] = pack(self.scales[0])   # DX
        label['X0V0'] = pack(self.offsets[0])  # X0
        label['NTITLE'] = len(title)
        label['TITLE'][:, :len(title)] = [ord(s) for s in title]
        xunit = self.units[0] if self.units[0] is not Undefined else ''
        label['XUNIT'][:, :len(xunit)] = [ord(c) for c in xunit]
        yunit = self.units[0] if self.units[0] is not Undefined else ''
        label['YUNIT'][:, :len(yunit)] = [ord(c) for c in yunit]
        zunit = self.units[0] if self.units[0] is not Undefined else ''
        label['ZUNIT'][:, :len(zunit)] = [ord(c) for c in zunit]
        return label

    @classmethod
    def _check_format(cls, data):
        if np.issubdtype(data.dtype, np.int8):
            iform = 0  # byte
        elif np.issubdtype(data.dtype, np.int16):
            iform = 1  # int16
        elif np.issubdtype(data.dtype, np.floating) and data.dtype.itemsize <= 4:
            data = data.astype(np.float32)
            iform = 2  # float (4 byte or less)
        elif np.issubdtype(data.dtype, np.complexfloating) and data.dtype.itemsize <= 8:
            data = data.astype(np.complex64)
            iform = 3  # complex (8 byte or less)
        elif np.issubdtype(data.dtype, np.int32):
            iform = 4  # int32
        else:
            supported_formats = [
                np.dtype(i).name for i in cls.IFORM_DICT.values()]
            msg = ('The SEMPER file format does not support '
                   '{} data type. '.format(data.dtype.name))
            msg += 'Supported data types are: ' + ', '.join(supported_formats)
            raise IOError(msg)
        return data, iform

    @classmethod
    def load_from_unf(cls, filename, lazy=False):
        r"""Load a `.unf`-file into a :class:`~.SemperFormat` object.

        Parameters
        ----------
        filename : string
            The name of the unf-file from which to load the data. Standard
            format is '\*.unf'.

        Returns
        -------
        semper : :class:`~.SemperFormat` (N=1)
            SEMPER file format object containing the loaded information.

        """
        metadata = OrderedDict()
        with open(filename, 'rb') as f:
            # Read header:
            rec_length = np.fromfile(
                f,
                dtype='<i4',
                count=1)[0]  # length of header
            header = np.fromfile(
                f,
                dtype=cls.HEADER_DTYPES[
                    :rec_length //
                    2],
                count=1)
            metadata.update(sarray2dict(header))
            assert np.frombuffer(f.read(4), dtype=np.int32)[0] == rec_length, \
                'Error while reading the header (length is not correct)!'
            data_format = cls.IFORM_DICT[metadata['IFORM']]
            iversn, remain = divmod(metadata['IFLAG'], 10000)
            ilabel, ntitle = divmod(remain, 1000)
            metadata.update(
                {'IVERSN': iversn, 'ILABEL': ilabel, 'NTITLE': ntitle})
            # Read title:
            title = ''
            if ntitle > 0:
                assert np.fromfile(
                    f,
                    dtype='<i4',
                    count=1)[0] == ntitle  # length of title
                title = b''.join(np.fromfile(f, dtype='c', count=ntitle))
                title = title.decode()
                metadata['TITLE'] = title
                assert np.fromfile(f, dtype='<i4', count=1)[0] == ntitle
            if ilabel:
                try:
                    metadata.update(cls._read_label(f))
                except Exception as e:
                    warning = ('Could not read label, trying to proceed '
                               'without it!')
                    warning += ' (Error message: {})'.format(str(e))
                    warnings.warn(warning)
            # Read picture data:
            pos = f.tell()
            shape = metadata['NLAY'], metadata['NROW'], metadata['NCOL']
            if lazy:
                from dask.array import from_delayed
                from dask import delayed
                task = delayed(_read_data)(f, filename, pos, data_format,
                                           shape)
                data = from_delayed(task, shape=shape, dtype=data_format)
            else:
                data = _read_data(f, filename, pos, data_format, shape)
        offsets = (metadata.get('X0V0', 0.),
                   metadata.get('Y0V2', 0.),
                   metadata.get('Z0V4', 0.))
        scales = (metadata.get('DXV1', 1.),
                  metadata.get('DYV3', 1.),
                  metadata.get('DZV5', 1.))
        units = (metadata.get('XUNIT', Undefined),
                 metadata.get('YUNIT', Undefined),
                 metadata.get('ZUNIT', Undefined))
        return cls(data, title, offsets, scales, units, metadata)

    def save_to_unf(self, filename='semper.unf', skip_header=False):
        """Save a :class:`~.SemperFormat` to a file.

        Parameters
        ----------
        filename : string, optional
            The name of the unf-file to which the data should be written.
        skip_header : bool, optional
            Determines if the header, title and label should be skipped (useful
            for some other programs). Default is False.

        Returns
        -------
        None

        """
        nlay, nrow, ncol = self.data.shape
        data, iform = self._check_format(self.data)
        title = self.title if self.title is not Undefined else ''
        with open(filename, 'wb') as f:
            if not skip_header:
                # Create header:
                header = np.zeros(
                    (1,), dtype=self.HEADER_DTYPES[
                        :-1])  # IFORMAT is not used!
                # Fill header:
                header['NCOL'] = self.data.shape[2]
                header['NROW'] = self.data.shape[1]
                header['NLAY'] = self.data.shape[0]
                header['ICLASS'] = self.metadata.get(
                    'ICLASS',
                    6)  # 6: Undefined!
                header['IFORM'] = iform
                # IVERSN: 2; ILABEL:1 (True)
                header['IFLAG'] = 2 * 10000 + 1000 + len(title)
                # Write header:
                f.write(
                    struct.pack(
                        '<i',
                        12))  # record length, 4 byte format!
                f.write(header.tobytes())
                f.write(
                    struct.pack(
                        '<i',
                        12))  # record length, 4 byte format!
                # Write title:
                if len(title) > 0:
                    f.write(
                        struct.pack(
                            '<i',
                            len(title)))  # record length, 4 byte format!
                    f.write(title.encode())
                    f.write(
                        struct.pack(
                            '<i',
                            len(title)))  # record length, 4 byte format!
                # Create label:
                label = self._get_label()
                # Write label:
                f.write(
                    struct.pack(
                        '<i',
                        2 *
                        256))  # record length, 4 byte format!
                f.write(label.tobytes())
                f.write(
                    struct.pack(
                        '<i',
                        2 *
                        256))  # record length, 4 byte format!
            # Write picture data:
            for k in range(nlay):
                for j in range(nrow):
                    row = self.data[k, j, :]
                    # Record length = bytes per entry * ncol:
                    record_length = np.dtype(
                        self.IFORM_DICT[iform]).itemsize * ncol
                    f.write(
                        struct.pack(
                            '<i',
                            record_length))  # record length, 4 byte format!
                    f.write(row.tobytes())
                    # SEMPER always expects an even number of bytes per row,
                    # which is only a problem for writing single byte data
                    # (IFORM = 0, np.byte). If ncol is odd, an empty byte (0)
                    # is added:
                    if self.data.dtype == np.byte and ncol % 2 != 0:
                        np.zeros(1, dtype=np.byte).tobytes()
                    # record length, 4 byte format!
                    f.write(struct.pack('<i', record_length))

    @classmethod
    def from_signal(cls, signal):
        """Import a :class:`~.SemperFormat` object from a
        :class:`~hyperspy.signals.Signal` object.

        Parameters
        ----------
        signal: :class:`~hyperspy.signals.Signal`
            The signal which should be imported.

        Returns
        -------
        None

        """
        data = signal.data
        assert len(data.shape) <= 3, \
            'Only up to 3-dimensional datasets can be handled!'
        scales, offsets, units = [1.] * 3, [0.] * \
            3, [Undefined] * 3  # Defaults!
        for i in range(len(data.shape)):
            scales[i] = signal.axes_manager[i].scale
            offsets[i] = signal.axes_manager[i].offset
            units[i] = signal.axes_manager[i].units
        # Make sure data is 3D!
        data = data[tuple(None for _ in range(3 - len(data.shape)))]
        signal_dimension = signal.axes_manager.signal_dimension
        if signal_dimension == 1:
            record_by = "spectrum"
        elif signal_dimension == 2:
            record_by = "image"
        else:
            record_by = ""
        iclass = cls.ICLASS_DICT_INV.get(record_by, 6)  # 6: undefined
        data, iform = cls._check_format(data)
        title = signal.metadata.General.as_dictionary().get('title', Undefined)
        metadata = OrderedDict()
        if 'date' in signal.metadata.General.keys(
        ) and 'time' in signal.metadata.General.keys():
            dt = "%s %s" % (signal.metadata.General.date,
                            signal.metadata.General.time)
        else:
            dt = "%s" % datetime.now()

        metadata.update({'DATE': "%s" % dt.split('.')[0],
                         'ICLASS': iclass,
                         'IFORM': iform,
                         'IVERSN': 2,  # Current standard
                         'ILABEL': 1,  # True
                         'IWP': 0,  # Seems standard
                         'IPLTYP': 248,  # Seems standard
                         'ICCOLN': data.shape[2] // 2 + 1,
                         'ICROWN': data.shape[1] // 2 + 1,
                         'ICLAYN': data.shape[0] // 2 + 1})
        return cls(data, title, offsets, scales, units, metadata)

    def to_signal(self, lazy=False):
        """Export a :class:`~.SemperFormat` object to a
        :class:`~hyperspy.signals.Signal` object.

        Parameters
        ----------
        None

        Returns
        -------
        signal: :class:`~hyperspy.signals.Signal`
            The exported signal.

        """
        import hyperspy.api as hp
        data = self.data.squeeze()  # Reduce unneeded dimensions!
        iclass = self.ICLASS_DICT.get(self.metadata.get('ICLASS'))
        if iclass == 'spectrum':
            signal = hp.signals.Signal1D(data)
        elif iclass == 'image':
            signal = hp.signals.Signal2D(data)
        else:  # Class is not given, but can be determined by the data shape:
            if len(data.shape) == 1:
                signal = hp.signals.Signal1D(data)
            elif len(data.shape) == 2:
                signal = hp.signals.Signal2D(data)
            else:  # 3D data!
                signal = hp.signals.BaseSignal(data)
        for i in range(len(data.shape)):
            signal.axes_manager[i].name = {0: 'x', 1: 'y', 2: 'z'}[i]
            signal.axes_manager[i].scale = self.scales[i]
            signal.axes_manager[i].offset = self.offsets[i]
            signal.axes_manager[i].units = self.units[i]
        signal.metadata.set_item('General.title', self.title)
        if 'DATE' in self.metadata.keys():
            date, time = self._convert_date_time_from_label()
            signal.metadata.set_item('General.date', date)
            signal.metadata.set_item('General.time', time)
        if lazy:
            signal = signal.as_lazy()
        signal.original_metadata.add_dictionary(self.metadata)
        return signal

    def _convert_date_time_from_label(self):
        # Convert the label['DATE'] to ISO 8601 for metadata
        try:
            dt = datetime.strptime(self.metadata['DATE'], "%Y-%m-%d %H:%M:%S")
        except ValueError:
            dt = datetime.strptime(self.metadata['DATE'], "%y-%m-%d %H:%M:%S")
        return dt.date().isoformat(), dt.time().isoformat()

    def log_info(self):
        """log important flag information of the :class:`.~SemperFormat`
        object.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        info_str = (
            "Semper info:\n" +
            ("\t%s\n" % self.title) +
            ('\tdimensions: x: {}, y: {}, z: {}\n'.format(
                *reversed(self.data.shape))) +
            ('\tscaling:    x: {:.3g}, y: {:.3g}, z: {:.3g}\n'.format(
                *self.scales)) +
            ('\toffsets:    x: {:.3g}, y: {:.3g}, z: {:.3g}'.format(
                *self.offsets)) +
            ('\tunits:      x: {}, y: {}, z: {}'.format(*self.units)) +
            ('\tdata range: %s %s\n' % (self.data.min(), self.data.max())) +
            ('\tmetadata:\n'))
        for k, v in self.metadata.items():
            info_str += '\t\t{}: {}'.format(k, v)
        _logger.info(info_str)


def unpack_from_intbytes(fmt, byte_list):
    """Read in a list of bytes (as int with range 0-255) and unpack them with
    format `fmt`.
    """
    return struct.unpack(fmt, b''.join(
        map(bytes, [[byte] for byte in byte_list])))[0]


def pack_to_intbytes(fmt, value):
    """Pack a `value` into a byte list using format `fmt` and represent it as
    int (range 0-255).
    """
    return [int(c) for c in struct.pack(fmt, value)]


def _read_data(fobj, fname, position, data_format, shape):
    if fobj.closed:
        fobj = open(fname, 'rb')
    fobj.seek(position)
    nlay, nrow, ncol = shape
    data = np.empty(shape, dtype=data_format)
    for k in range(nlay):
        for j in range(nrow):
            rec_length = np.fromfile(fobj, dtype='<i4', count=1)[0]
            # Not always ncol, see below
            count = rec_length // np.dtype(data_format).itemsize
            row = np.fromfile(fobj, dtype=data_format, count=count)
            # [:ncol] is used because Semper always writes an even
            # number of bytes which is a problem when reading in single
            # bytes (IFORM = 0, np.byte). If ncol is odd, an empty
            # byte (0) is added which has to be skipped during read in:
            data[k, j, :] = row[:ncol]
            test = np.fromfile(fobj, dtype='<i4', count=1)[0]
            assert test == rec_length
    return data


def file_reader(filename, **kwds):
    lazy = kwds.get('lazy', False)
    semper = SemperFormat.load_from_unf(filename, lazy=lazy)
    semper.log_info()
    return [semper.to_signal(lazy=lazy)._to_dictionary()]


def file_writer(filename, signal, **kwds):
    semper = SemperFormat.from_signal(signal)
    semper.save_to_unf(filename)



import tempfile

import dask.array as da
import h5py
import numpy as np
import pytest

from hyperspy import api as hs

usid = pytest.importorskip("pyUSID", reason="pyUSID not installed")
sidpy = pytest.importorskip("sidpy", reason="sidpy not installed")


# ##################### HELPER FUNCTIONS ######################################

def _array_translator_basic_checks(h5_f):
    assert len(h5_f.items()) == 1
    assert 'Measurement_000' in h5_f.keys()
    h5_meas_grp = h5_f['Measurement_000']
    assert isinstance(h5_meas_grp, h5py.Group)

    # Again, this group should only have one group - Channel_000
    assert len(h5_meas_grp.items()) == 1
    assert 'Channel_000' in h5_meas_grp.keys()
    h5_chan_grp = h5_meas_grp['Channel_000']
    assert isinstance(h5_chan_grp, h5py.Group)

    # This channel group will contain the main dataset
    assert len(h5_chan_grp.items()) == 5
    for dset_name in ['Raw_Data', 'Position_Indices', 'Position_Values',
                      'Spectroscopic_Indices', 'Spectroscopic_Values']:
        assert dset_name in h5_chan_grp.keys()
        h5_dset = h5_chan_grp[dset_name]
        assert isinstance(h5_dset, h5py.Dataset)

    assert len(usid.hdf_utils.get_all_main(h5_f)) == 1

    assert usid.hdf_utils.check_if_main(h5_chan_grp['Raw_Data'])


def _compare_axes(hs_axes, dim_descriptors, usid_val_func, axes_defined=True,
                  invalid_axes=False):
    usid_descriptors = [_split_descriptor(_) for _ in dim_descriptors]
    for dim_ind, axis in enumerate(hs_axes):
        # Remember that the signal axes are actually reversed!
        real_dim_ind = len(hs_axes) - dim_ind - 1
        if axes_defined:
            assert axis.name == usid_descriptors[real_dim_ind][0]
            assert axis.units == usid_descriptors[real_dim_ind][1]

        if not invalid_axes:
            axis_vals = np.arange(axis.offset,
                                  axis.offset + axis.size * axis.scale,
                                  axis.scale)
            usid_vals = usid_val_func(usid_descriptors[real_dim_ind][0])
            np.testing.assert_allclose(axis_vals, usid_vals)


def _assert_empty_dims(hs_axes, usid_labels, usid_val_func):
    assert len(hs_axes) == 0
    assert len(usid_labels) == 1
    dim_vals = usid_val_func(usid_labels[0])
    assert len(dim_vals) == 1


def _split_descriptor(desc):
    desc = desc.strip()
    ind = desc.rfind('(')
    if ind < 0:
        ind = desc.rfind('[')
        if ind < 0:
            return desc, ''

    quant = desc[:ind].strip()
    units = desc[ind:]
    for item in '()[]':
        units = units.replace(item, '')
    return quant, units


def _validate_metadata_from_h5dset(sig, h5_dset, compound_comp_name=None):
    if compound_comp_name is None:
        quant = usid.hdf_utils.get_attr(h5_dset, 'quantity')
        units = usid.hdf_utils.get_attr(h5_dset, 'units')
    else:
        quant, units = _split_descriptor(compound_comp_name)
    assert sig.original_metadata.quantity == quant
    assert sig.original_metadata.units == units
    assert sig.metadata.General.original_filename == h5_dset.file.filename
    assert sig.metadata.General.title == h5_dset.name.split('/')[-1]
    assert sig.original_metadata.dataset_path == h5_dset.name
    assert sig.original_metadata.original_file_type == 'USID HDF5'
    assert sig.original_metadata.pyUSID_version == usid.__version__

    h5_chan_grp = h5_dset.parent
    usid_grp_parms = dict()
    if 'Channel' in h5_chan_grp.name.split('/')[-1]:
        usid_grp_parms = sidpy.hdf.hdf_utils.get_attributes(h5_chan_grp)
        h5_meas_grp = h5_chan_grp.parent
        if 'Measurement' in h5_meas_grp.name.split('/')[-1]:
            temp = sidpy.hdf.hdf_utils.get_attributes(h5_meas_grp)
            usid_grp_parms.update(temp)
    # Remove timestamp key since there is 1s difference occasionally
    usid_grp_parms.pop('timestamp', None)
    om_dict = sig.original_metadata.parameters.as_dictionary()
    om_dict.pop('timestamp', None)
    assert om_dict == usid_grp_parms


def compare_usid_from_signal(sig, h5_path, empty_pos=False, empty_spec=False,
                             dataset_path=None, **kwargs):
    with h5py.File(h5_path, mode='r') as h5_f:
        # 1. Validate that what has been written is a USID Main dataset
        if dataset_path is None:
            _array_translator_basic_checks(h5_f)
            h5_main = usid.hdf_utils.get_all_main(h5_f)[0]
        else:
            h5_main = usid.USIDataset(h5_f[dataset_path])

        usid_data = h5_main.get_n_dim_form().squeeze()
        # 2. Validate that raw data has been written correctly:
        np.testing.assert_allclose(sig.data, usid_data)
        # 3. Validate that axes / dimensions have been translated correctly:
        if empty_pos:
            _assert_empty_dims(sig.axes_manager.navigation_axes,
                               h5_main.pos_dim_labels, h5_main.get_pos_values,
                               **kwargs)
        else:
            _compare_axes(sig.axes_manager.navigation_axes,
                          h5_main.pos_dim_descriptors, h5_main.get_pos_values,
                          **kwargs)
        # 4. Check to make sure that there is only one spectroscopic dimension
        # of size 1
        if empty_spec:
            _assert_empty_dims(sig.axes_manager.signal_axes,
                               h5_main.spec_dim_labels,
                               h5_main.get_spec_values, **kwargs)
        else:
            _compare_axes(sig.axes_manager.signal_axes,
                          h5_main.spec_dim_descriptors,
                          h5_main.get_spec_values, **kwargs)


def compare_signal_from_usid(file_path, ndata, new_sig, axes_to_spec=[],
                             sig_type=hs.signals.BaseSignal, dataset_path=None,
                             compound_comp_name=None, **kwargs):
    # 1. Validate object type
    assert isinstance(new_sig, sig_type)
    if len(axes_to_spec) > 0:
        new_sig = new_sig.as_signal2D(axes_to_spec)

    # 2. Validate that data has been read in correctly:
    np.testing.assert_allclose(new_sig.data, ndata)
    with h5py.File(file_path, mode='r') as h5_f:
        if dataset_path is None:
            h5_main = usid.hdf_utils.get_all_main(h5_f)[0]
        else:
            h5_main = usid.USIDataset(h5_f[dataset_path])
        # 3. Validate that all axes / dimensions have been translated correctly
        if len(axes_to_spec) > 0:
            _compare_axes(new_sig.axes_manager.navigation_axes,
                          h5_main.pos_dim_descriptors, h5_main.get_pos_values,
                          **kwargs)
        else:
            assert new_sig.axes_manager.navigation_dimension == 0
        # 3. Validate that all spectroscopic axes / dimensions have been
        # translated correctly:
        if h5_main.shape[1] == 1:
            _compare_axes(new_sig.axes_manager.signal_axes,
                          h5_main.pos_dim_descriptors, h5_main.get_pos_values,
                          **kwargs)
        else:
            _compare_axes(new_sig.axes_manager.signal_axes,
                          h5_main.spec_dim_descriptors,
                          h5_main.get_spec_values, **kwargs)

        # 5. Validate that metadata has been read in correctly:
        _validate_metadata_from_h5dset(new_sig, h5_main,
                                       compound_comp_name=compound_comp_name)


def gen_2pos_2spec(s2f_aux=True, mode=None):
    pos_dims = [usid.Dimension('X', 'nm', [-250, 750]),
                usid.Dimension('Y', 'um', np.linspace(0, 60, num=7))]
    spec_dims = [usid.Dimension('Frequency', 'kHz', [300, 350, 400]),
                 usid.Dimension('Bias', 'V', np.linspace(-4, 4, num=5))]
    if s2f_aux:
        pos_dims = pos_dims[::-1]
        spec_dims = spec_dims[::-1]

    ndim_shape = (7, 2, 5, 3)
    if mode is None:
        # Typcial floating point dataset
        ndata = np.random.rand(*ndim_shape)
    elif mode == 'complex':
        ndata = np.random.rand(*ndim_shape) + 1j * np.random.rand(*ndim_shape)
    elif mode == 'compound':
        struc_dtype = np.dtype({'names': ['amp', 'phas'],
                                'formats': [np.float16, np.float32]})
        ndata = np.zeros(shape=ndim_shape, dtype=struc_dtype)
        ndata['amp'] = np.random.random(size=ndim_shape)
        ndata['phas'] = np.random.random(size=ndim_shape)

    data_2d = ndata.reshape(np.prod(ndata.shape[:2]),
                              np.prod(ndata.shape[2:]))
    return pos_dims, spec_dims, ndata, data_2d


def gen_2dim(all_pos=False, s2f_aux=True):
    ndata = np.random.rand(3, 2)
    if all_pos:
        pos_dims = [usid.Dimension('X', 'nm', [-250, 750]),
                    usid.Dimension('Y', 'um', [-2, 0, 2])]
        spec_dims = [usid.Dimension('arb', 'a.u.', 1)]
        data_2d = ndata.reshape(-1, 1)
    else:
        pos_dims = [usid.Dimension('arb', 'a.u.', 1)]
        spec_dims = [usid.Dimension('Frequency', 'kHz', [0, 100]),
                     usid.Dimension('Bias', 'V', [-5, 0, 5])]
        data_2d = ndata.reshape(1, -1)
    if s2f_aux:
        pos_dims = pos_dims[::-1]
        spec_dims = spec_dims[::-1]
    return pos_dims, spec_dims, ndata, data_2d

# ################ HyperSpy Signal to h5USID ##################################


@pytest.mark.filterwarnings("ignore:This dataset does not have an N-dimensional form:UserWarning")
class TestHS2USIDallKnown:

    def test_n_pos_0_spec(self):
        sig = hs.signals.BaseSignal(np.random.randint(0, high=100,
                                                      size=(2, 3)),
                                    axes=[{'size': 2, 'name': 'X',
                                           'units': 'nm', 'scale': 0.2,
                                           'offset': 1},
                                          {'size': 3, 'name': 'Y',
                                           'units': 'um', 'scale': 0.1,
                                           'offset': 2}])
        sig = sig.transpose()

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_0_spec_real.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=False,
                                 empty_spec=True)

    def test_0_pos_n_spec(self):
        sig = hs.signals.BaseSignal(np.random.randint(0, high=100,
                                                      size=(2, 3)),
                                    axes=[{'size': 2, 'name': 'Freq',
                                           'units': 'Hz', 'scale': 30,
                                           'offset': 300},
                                          {'size': 3, 'name': 'Bias',
                                           'units': 'V', 'scale': 0.25,
                                           'offset': -0.25}])

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_0_pos_n_spec_real.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=True,
                                 empty_spec=False)

    def test_n_pos_m_spec(self):
        sig = hs.signals.Signal2D(np.random.randint(0, high=100,
                                                    size=(2, 3, 5, 7)),
                                  axes=[{'size': 2, 'name': 'X', 'units': 'nm',
                                         'scale': 0.2, 'offset': 1},
                                        {'size': 3, 'name': 'Y', 'units': 'um',
                                         'scale': 0.1, 'offset': 2},
                                        {'size': 5, 'name': 'Freq',
                                         'units': 'Hz', 'scale': 30,
                                         'offset': 300},
                                        {'size': 7, 'name': 'Bias',
                                         'units': 'V', 'scale': 0.25,
                                         'offset': -0.25}])

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_real.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=False,
                                 empty_spec=False)

    def test_append_to_existing_file(self):
        sig = hs.signals.BaseSignal(np.random.randint(0, high=100,
                                                      size=(2, 3)),
                                    axes=[{'size': 2, 'name': 'X',
                                           'units': 'nm', 'scale': 0.2,
                                           'offset': 1},
                                          {'size': 3, 'name': 'Y',
                                           'units': 'um', 'scale': 0.1,
                                           'offset': 2}])
        sig = sig.transpose()

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_0_spec_real.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=False,
                                 empty_spec=True)

        sig = hs.signals.BaseSignal(np.random.randint(0, high=100,
                                                      size=(2, 3)),
                                    axes=[{'size': 2, 'name': 'Freq',
                                           'units': 'Hz', 'scale': 30,
                                           'offset': 300},
                                          {'size': 3, 'name': 'Bias',
                                           'units': 'V', 'scale': 0.25,
                                           'offset': -0.25}])

        sig.save(file_path, overwrite=True)

        new_dataset_path = '/Measurement_001/Channel_000/Raw_Data'

        compare_usid_from_signal(sig, file_path, empty_pos=True,
                                 empty_spec=False, dataset_path=new_dataset_path)


@pytest.mark.filterwarnings("ignore:This dataset does not have an N-dimensional form:UserWarning")
class TestHS2USIDlazy:

    def test_base_nd(self):
        sig = hs.signals.Signal2D(da.random.randint(0, high=100,
                                                    size=(2, 3, 5, 7),
                                                    chunks='auto')).as_lazy()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_dask.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=False,
                                 empty_spec=False, axes_defined=False)


@pytest.mark.filterwarnings("ignore:This dataset does not have an N-dimensional form:UserWarning")
class TestHS2USIDedgeAxes:

    def test_no_axes(self):
        sig = hs.signals.Signal2D(np.random.randint(0, high=100,
                                                    size=(2, 3, 5, 7)))

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_no_axes.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=False,
                                 empty_spec=False, axes_defined=False)

    def test_incorrect_axes(self):
        sig = hs.signals.Signal2D(np.random.randint(0, high=100,
                                                    size=(2, 3, 5, 7)),
                                  axes=[{'size': 1259, 'name': 'X',
                                         'units': 'nm', 'scale': 0.2,
                                         'offset': 1},
                                        {'size': 245, 'name': 'Y',
                                         'units': 'um', 'scale': 0.1,
                                         'offset': 2},
                                        {'size': 205, 'name': 'Freq',
                                         'units': 'Hz', 'scale': 30,
                                         'offset': 300},
                                        {'size': 1005, 'name': 'Bias',
                                         'units': 'V', 'scale': 0.25,
                                         'offset': -0.25}])

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_incorrect_axes.h5'
        sig.save(file_path)

        compare_usid_from_signal(sig, file_path, empty_pos=False,
                                 empty_spec=False, invalid_axes=True)

    def test_too_many_axes(self):

        sig = hs.signals.Signal2D(np.random.randint(0, high=100,
                                                    size=(2, 3, 5, 7)),
                                  axes=[{'size': 2, 'name': 'X', 'units': 'nm',
                                         'scale': 0.2, 'offset': 1},
                                        {'size': 3, 'name': 'Y', 'units': 'um',
                                         'scale': 0.1, 'offset': 2},
                                        {'size': 5, 'name': 'Freq',
                                         'units': 'Hz', 'scale': 30,
                                         'offset': 300},
                                        {'size': 7, 'name': 'Bias',
                                         'units': 'V', 'scale': 0.25,
                                         'offset': -0.25},
                                        {'size': 27, 'name': 'Does not exist',
                                         'units': 'V', 'scale': 0.25,
                                         'offset': -0.25}])

        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_too_many_axes.h5'
        with pytest.raises(ValueError):
            sig.save(file_path)

# ################## h5USID to HyperSpy Signal(s)  ############################

@pytest.mark.filterwarnings("ignore:This dataset does not have an N-dimensional form:UserWarning")
class TestUSID2HSbase:

    def test_n_pos_0_spec(self):
        phy_quant = 'Current'
        phy_unit = 'nA'
        slow_to_fast = True
        pos_dims, spec_dims, ndata, data_2d = gen_2dim(all_pos=True,
                                                       s2f_aux=slow_to_fast)
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_0_spec.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        new_sig = hs.load(file_path)
        compare_signal_from_usid(file_path, ndata, new_sig,
                                 sig_type=hs.signals.BaseSignal,
                                 axes_to_spec=[])

    def test_0_pos_n_spec(self):
        phy_quant = 'Current'
        phy_unit = 'nA'
        slow_to_fast = True
        pos_dims, spec_dims, ndata, data_2d = gen_2dim(all_pos=False,
                                                       s2f_aux=slow_to_fast)
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_0_pos_n_spec.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        new_sig = hs.load(file_path)
        compare_signal_from_usid(file_path, ndata, new_sig,
                                 sig_type=hs.signals.BaseSignal,
                                 axes_to_spec=[])

    @pytest.mark.parametrize('lazy', [True, False])
    def test_base_n_pos_m_spec(self, lazy, slow_to_fast=True):
        phy_quant = 'Current'
        phy_unit = 'nA'
        ret_vals = gen_2pos_2spec(s2f_aux=slow_to_fast)
        pos_dims, spec_dims, ndata, data_2d = ret_vals
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        new_sig = hs.load(file_path, lazy=lazy)
        compare_signal_from_usid(file_path, ndata, new_sig,
                                 sig_type=hs.signals.BaseSignal,
                                 axes_to_spec=['Frequency', 'Bias'])


@pytest.mark.filterwarnings("ignore:This dataset does not have an N-dimensional form:UserWarning")
class TestUSID2HSdtype:

    def test_complex(self):
        phy_quant = 'Current'
        phy_unit = 'nA'
        slow_to_fast = True
        ret_vals = gen_2pos_2spec(s2f_aux=slow_to_fast, mode='complex')
        pos_dims, spec_dims, ndata, data_2d = ret_vals
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_complex.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        new_sig = hs.load(file_path)
        compare_signal_from_usid(file_path, ndata, new_sig,
                                 sig_type=hs.signals.ComplexSignal,
                                 axes_to_spec=['Frequency', 'Bias'])

    def test_compound(self):
        phy_quant = 'Current'
        phy_unit = 'nA'
        slow_to_fast = True
        ret_vals = gen_2pos_2spec(s2f_aux=slow_to_fast, mode='compound')
        pos_dims, spec_dims, ndata, data_2d = ret_vals
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_compound.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        objects = hs.load(file_path)
        assert isinstance(objects, list)
        assert len(objects) == 2

        # 1. Validate object type
        for new_sig, comp_name in zip(objects, ['amp', 'phas']):
            compare_signal_from_usid(file_path, ndata[comp_name], new_sig,
                                     sig_type=hs.signals.BaseSignal,
                                     axes_to_spec=['Frequency', 'Bias'],
                                     compound_comp_name=comp_name)

    def test_non_linear_dimension(self):
        pos_dims = [usid.Dimension('Y', 'um', np.linspace(0, 60, num=5)),
                    usid.Dimension('X', 'nm', [-250, 750])]
        spec_dims = [usid.Dimension('Bias', 'V',
                                    np.sin(np.linspace(0, 2 * np.pi, num=7))),
                     usid.Dimension('Frequency', 'kHz', [300, 350, 400])]
        ndata = np.random.rand(5, 2, 7, 3)
        phy_quant = 'Current'
        phy_unit = 'nA'
        data_2d = ndata.reshape(np.prod(ndata.shape[:2]),
                                np.prod(ndata.shape[2:]))
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec_non_lin_dim.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims,
                           slow_to_fast=True)

        with pytest.raises(ValueError):
            _ = hs.load(file_path, ignore_non_linear_dims=False)

        with pytest.warns(UserWarning) as _:
            new_sig = hs.load(file_path)
        compare_signal_from_usid(file_path, ndata, new_sig,
                                 axes_to_spec=['Frequency', 'Bias'],
                                 invalid_axes=True)


@pytest.mark.filterwarnings("ignore:This dataset does not have an N-dimensional form:UserWarning")
class TestUSID2HSmultiDsets:

    def test_pick_specific(self):
        slow_to_fast = True
        ret_vals = gen_2pos_2spec(s2f_aux=slow_to_fast)
        pos_dims, spec_dims, ndata, data_2d = ret_vals
        phy_quant = 'Current'
        phy_unit = 'nA'
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_n_pos_n_spec.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        ret_vals = gen_2dim(all_pos=True, s2f_aux=slow_to_fast)
        pos_dims, spec_dims, ndata_2, data_2d_2 = ret_vals
        phy_quant = 'Current'
        phy_unit = 'nA'

        with h5py.File(file_path, mode='r+') as h5_f:
            h5_meas_grp = h5_f.create_group('Measurement_001')
            h5_chan_grp = h5_meas_grp.create_group('Channel_000')
            _ = usid.hdf_utils.write_main_dataset(h5_chan_grp, data_2d_2,
                                                  'Raw_Data', phy_quant,
                                                  phy_unit, pos_dims,
                                                  spec_dims,
                                                  slow_to_fast=slow_to_fast)

        dataset_path = '/Measurement_001/Channel_000/Raw_Data'
        new_sig = hs.load(file_path, dataset_path=dataset_path)
        compare_signal_from_usid(file_path, ndata_2, new_sig,
                                 dataset_path=dataset_path)

    def test_read_all_by_default(self):
        slow_to_fast = True
        pos_dims, spec_dims, ndata, data_2d = gen_2dim(all_pos=False,
                                                       s2f_aux=slow_to_fast)
        phy_quant = 'Current'
        phy_unit = 'nA'
        tran = usid.ArrayTranslator()
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = tmp_dir + 'usid_0_pos_n_spec.h5'
        _ = tran.translate(file_path, 'Blah', data_2d, phy_quant, phy_unit,
                           pos_dims, spec_dims, slow_to_fast=slow_to_fast)

        ret_vals = gen_2dim(all_pos=True, s2f_aux=slow_to_fast)
        pos_dims, spec_dims, ndata_2, data_2d_2 = ret_vals
        phy_quant = 'Current'
        phy_unit = 'nA'

        with h5py.File(file_path, mode='r+') as h5_f:
            h5_meas_grp = h5_f.create_group('Measurement_001')
            _ = usid.hdf_utils.write_main_dataset(h5_meas_grp, data_2d_2,
                                                  'Spat_Map', phy_quant,
                                                  phy_unit, pos_dims,
                                                  spec_dims,
                                                  slow_to_fast=slow_to_fast)

        objects = hs.load(file_path)
        assert isinstance(objects, list)
        assert len(objects) == 2

        dset_names = ['/Measurement_000/Channel_000/Raw_Data',
                      'Measurement_001/Spat_Map']

        # 1. Validate object type
        for new_sig, ndim_data, dataset_path in zip(objects,
                                                    [ndata, ndata_2],
                                                    dset_names):
            compare_signal_from_usid(file_path, ndim_data, new_sig,
                                     dataset_path=dataset_path)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from hyperspy import utils
from hyperspy.signal import BaseSignal
from hyperspy.exceptions import VisibleDeprecationWarning


def test_stack_warning():
    with pytest.warns(VisibleDeprecationWarning, match="deprecated"):
        _ = utils.stack([BaseSignal([1]), BaseSignal([2])], mmap=True)


class TestUtilsStack:

    def setup_method(self, method):
        s = BaseSignal(np.random.random((3, 2, 5)))
        s.axes_manager.set_signal_dimension(1)
        s.axes_manager[0].name = "x"
        s.axes_manager[1].name = "y"
        s.axes_manager[2].name = "E"
        s.axes_manager[2].scale = 0.5
        s.metadata.General.title = 'test'
        self.signal = s

    def test_stack_default(self):
        s = self.signal
        s1 = s.deepcopy() + 1
        s2 = s.deepcopy() * 4
        test_axis = s.axes_manager[0].index_in_array
        result_signal = utils.stack([s, s1, s2])
        result_list = result_signal.split()
        assert test_axis == s.axes_manager[0].index_in_array
        assert len(result_list) == 3
        np.testing.assert_array_almost_equal(
            result_list[0].data, result_signal.inav[:, :, 0].data)

    def test_stack_number_of_parts(self):
        s = self.signal
        s1 = s.deepcopy() + 1
        s2 = s.deepcopy() * 4
        test_axis = s.axes_manager[0].index_in_array
        result_signal = utils.stack([s, s1, s2])
        result_list = result_signal.split(number_of_parts=3)
        assert test_axis == s.axes_manager[0].index_in_array
        assert len(result_list) == 3
        np.testing.assert_array_almost_equal(
            result_list[0].data, result_signal.inav[:, :, 0].data)

    def test_stack_of_stack(self):
        s = self.signal
        s1 = utils.stack([s] * 2)
        s2 = utils.stack([s1] * 3)
        s3 = s2.split()[0]
        s4 = s3.split()[0]
        np.testing.assert_array_almost_equal(s4.data, s.data)
        assert not hasattr(s4.original_metadata, 'stack_elements')
        assert s4.metadata.General.title == 'test'

    def test_stack_not_default(self):
        s = self.signal
        s1 = s.inav[:, :-1] + 1
        s2 = s.inav[:, ::2] * 4
        result_signal = utils.stack([s, s1, s2], axis=1)
        axis_size = s.axes_manager[1].size
        axs1 = s1.axes_manager[1].size
        axs2 = s2.axes_manager[1].size
        result_list = result_signal.split()
        assert len(result_list) == 3
        for rs in [result_signal, utils.stack([s, s1, s2], axis='y')]:
            np.testing.assert_array_almost_equal(
                result_list[0].data, rs.inav[:, :axis_size].data)
            np.testing.assert_array_almost_equal(
                s.data, rs.inav[:, :axis_size].data)
            np.testing.assert_array_almost_equal(
                s1.data, rs.inav[:, axis_size:axis_size + axs1].data)
            np.testing.assert_array_almost_equal(
                s2.data, rs.inav[:, axis_size + axs1:].data)

    def test_stack_bigger_than_ten(self):
        s = self.signal
        list_s = [s] * 12
        list_s.append(s.deepcopy() * 3)
        list_s[-1].metadata.General.title = 'test'
        s1 = utils.stack(list_s)
        res = s1.split()
        np.testing.assert_array_almost_equal(list_s[-1].data, res[-1].data)
        assert res[-1].metadata.General.title == 'test'

    def test_stack_broadcast_number(self):
        s = self.signal
        rs = utils.stack([5, s])
        np.testing.assert_array_equal(
            rs.inav[..., 0].data, 5 * np.ones((3, 2, 5)))

    def test_stack_broadcast_number_not_default(self):
        s = self.signal
        rs = utils.stack([5, s], axis='E')
        np.testing.assert_array_equal(rs.isig[0].data, 5 * np.ones((3, 2)))



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division

import matplotlib.pyplot as plt
from matplotlib.backend_bases import MouseEvent
import numpy as np

from hyperspy.drawing.utils import on_figure_window_close
from hyperspy.events import Events, Event


class WidgetBase(object):

    """Base class for interactive widgets/patches. A widget creates and
    maintains one or more matplotlib patches, and manages the interaction code
    so that the user can maniuplate it on the fly.

    This base class implements functionality witch is common to all such
    widgets, mainly the code that manages the patch, axes management, and
    sets up common events ('changed' and 'closed').

    Any inherting subclasses must implement the following methods:
        _set_patch(self)
        _on_navigate(obj, name, old, new)  # Only for widgets that can navigate

    It should also make sure to fill the 'axes' attribute as early as
    possible (but after the base class init), so that it is available when
    needed.
    """

    def __init__(self, axes_manager=None, color='red', alpha=1.0, **kwargs):
        self.axes_manager = axes_manager
        self._axes = list()
        self.ax = None
        self.picked = False
        self.selected = False
        self._selected_artist = None
        self._size = 1.
        self._pos = np.array([0.])
        self.__is_on = True
        self.background = None
        self.patch = []
        self.color = color
        self.alpha = alpha
        self.cids = list()
        self.blit = True
        self.events = Events()
        self.events.changed = Event(doc="""
            Event that triggers when the widget has a significant change.

            The event triggers after the internal state of the widget has been
            updated.

            Arguments:
            ----------
                widget:
                    The widget that changed
            """, arguments=['obj'])
        self.events.closed = Event(doc="""
            Event that triggers when the widget closed.

            The event triggers after the widget has already been closed.

            Arguments:
            ----------
                widget:
                    The widget that closed
            """, arguments=['obj'])
        self._navigating = False
        super(WidgetBase, self).__init__(**kwargs)

    def _get_axes(self):
        return self._axes

    def _set_axes(self, axes):
        if axes is None:
            self._axes = list()
        else:
            self._axes = axes

    axes = property(lambda s: s._get_axes(),
                    lambda s, v: s._set_axes(v))

    def is_on(self):
        """Determines if the widget is set to draw if valid (turned on).
        """
        return self.__is_on

    def set_on(self, value):
        """Change the on state of the widget. If turning off, all patches will
        be removed from the matplotlib axes and the widget will disconnect from
        all events. If turning on, the patch(es) will be added to the
        matplotlib axes, and the widget will connect to its default events.
        """
        did_something = False
        if value is not self.is_on() and self.ax is not None:
            did_something = True
            if value is True:
                self._add_patch_to(self.ax)
                self.connect(self.ax)
            elif value is False:
                for container in [
                        self.ax.patches,
                        self.ax.lines,
                        self.ax.artists,
                        self.ax.texts]:
                    for p in self.patch:
                        if p in container:
                            container.remove(p)
                self.disconnect()
        if hasattr(super(WidgetBase, self), 'set_on'):
            super(WidgetBase, self).set_on(value)
        if did_something:
            self.draw_patch()
            if value is False:
                self.ax = None
        self.__is_on = value

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = color
        for p in self.patch:
            p.set_color(self._color)

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        self._alpha = alpha
        for p in self.patch:
            p.set_alpha(self._alpha)

    def _set_patch(self):
        """Create the matplotlib patch(es), and store it in self.patch
        """
        if hasattr(super(WidgetBase, self), '_set_patch'):
            super(WidgetBase, self)._set_patch()
        # Must be provided by the subclass

    def _add_patch_to(self, ax):
        """Create and add the matplotlib patches to 'ax'
        """
        self._set_patch()
        self.blit = hasattr(ax, 'hspy_fig') and ax.figure.canvas.supports_blit
        for p in self.patch:
            ax.add_artist(p)
            p.set_animated(self.blit)
        if hasattr(super(WidgetBase, self), '_add_patch_to'):
            super(WidgetBase, self)._add_patch_to(ax)

    def set_mpl_ax(self, ax):
        """Set the matplotlib Axes that the widget will draw to. If the widget
        on state is True, it will also add the patch to the Axes, and connect
        to its default events.
        """
        if ax is self.ax:
            return  # Do nothing
        # Disconnect from previous axes if set
        if self.ax is not None and self.is_on():
            self.disconnect()
        self.ax = ax
        canvas = ax.figure.canvas
        if self.is_on() is True:
            self._add_patch_to(ax)
            self.connect(ax)
            canvas.draw_idle()
            self.select()

    def select(self):
        """
        Cause this widget to be the selected widget in its MPL axes. This
        assumes that the widget has its patch added to the MPL axes.
        """
        if not self.patch or not self.is_on() or not self.ax:
            return

        canvas = self.ax.figure.canvas
        # Simulate a pick event
        x, y = self.patch[0].get_transform().transform_point((0, 0))
        mouseevent = MouseEvent('pick_event', canvas, x, y)
        # when the widget is added programatically, mouseevent can be "empty"
        if mouseevent.button:
            canvas.pick_event(mouseevent, self.patch[0])
        self.picked = False

    def connect(self, ax):
        """Connect to the matplotlib Axes' events.
        """
        on_figure_window_close(ax.figure, self.close)
        if self._navigating:
            self.connect_navigate()

    def connect_navigate(self):
        """Connect to the axes_manager such that changes in the widget or in
        the axes_manager are reflected in the other.
        """
        if self._navigating:
            self.disconnect_navigate()
        self.axes_manager.events.indices_changed.connect(
            self._on_navigate, {'obj': 'axes_manager'})
        self._on_navigate(self.axes_manager)    # Update our position
        self._navigating = True

    def disconnect_navigate(self):
        """Disconnect a previous naivgation connection.
        """
        self.axes_manager.events.indices_changed.disconnect(self._on_navigate)
        self._navigating = False

    def _on_navigate(self, axes_manager):
        """Callback for axes_manager's change notification.
        """
        pass    # Implement in subclass!

    def disconnect(self):
        """Disconnect from all events (both matplotlib and navigation).
        """
        for cid in self.cids:
            try:
                self.ax.figure.canvas.mpl_disconnect(cid)
            except BaseException:
                pass
        if self._navigating:
            self.disconnect_navigate()

    def close(self, window=None):
        """Set the on state to off (removes patch and disconnects), and trigger
        events.closed.
        """
        self.set_on(False)
        self.events.closed.trigger(obj=self)

    def draw_patch(self, *args):
        """Update the patch drawing.
        """
        try:
            if self.blit and hasattr(self.ax, 'hspy_fig'):
                self.ax.hspy_fig._update_animated()
            elif self.ax.figure is not None:
                self.ax.figure.canvas.draw_idle()
        except AttributeError:
            pass  # When figure is None, typically when closing

    def _v2i(self, axis, v):
        """Wrapped version of DataAxis.value2index, which bounds the index
        inbetween axis.low_index and axis.high_index+1, and does not raise a
        ValueError.
        """
        try:
            return axis.value2index(v)
        except ValueError:
            if v > axis.high_value:
                return axis.high_index + 1
            elif v < axis.low_value:
                return axis.low_index
            else:
                raise

    def _i2v(self, axis, i):
        """Wrapped version of DataAxis.index2value, which bounds the value
        inbetween axis.low_value and axis.high_value+axis.scale, and does not
        raise a ValueError.
        """
        try:
            return axis.index2value(i)
        except ValueError:
            if i > axis.high_index:
                return axis.high_value + axis.scale
            elif i < axis.low_index:
                return axis.low_value
            else:
                raise

    def __str__(self):
        return "{} with id {}".format(self.__class__.__name__, id(self))


class DraggableWidgetBase(WidgetBase):

    """Adds the `position` and `indices` properties, and adds a framework for
    letting the user drag the patch around. Also adds the `moved` event.

    The default behavior is that `position` snaps to the values corresponding
    to the values of the axes grid (i.e. no subpixel values). This behavior
    can be controlled by the property `snap_position`.

    Any inheritors must override these methods:
        _onmousemove(self, event)
        _update_patch_position(self)
        _set_patch(self)
    """

    def __init__(self, axes_manager, **kwargs):
        super(DraggableWidgetBase, self).__init__(axes_manager, **kwargs)
        self.events.moved = Event(doc="""
            Event that triggers when the widget was moved.

            The event triggers after the internal state of the widget has been
            updated. This event does not differentiate on how the position of
            the widget was changed, so it is the responsibility of the user
            to suppress events as neccessary to avoid closed loops etc.

            Arguments:
            ----------
                obj:
                    The widget that was moved.
            """, arguments=['obj'])
        self._snap_position = True

        # Set default axes
        if self.axes_manager is not None:
            if self.axes_manager.navigation_dimension > 0:
                self.axes = self.axes_manager.navigation_axes[0:1]
            else:
                self.axes = self.axes_manager.signal_axes[0:1]
        else:
            self._pos = np.array([0.])

    def _set_axes(self, axes):
        super(DraggableWidgetBase, self)._set_axes(axes)
        if self.axes:
            self._pos = np.array([ax.low_value for ax in self.axes])

    def _get_indices(self):
        """Returns a tuple with the position (indices).
        """
        idx = []
        for i in range(len(self.axes)):
            idx.append(self.axes[i].value2index(self._pos[i]))
        return tuple(idx)

    def _set_indices(self, value):
        """Sets the position of the widget (by indices). The dimensions should
        correspond to that of the 'axes' attribute. Calls _pos_changed if the
        value has changed, which is then responsible for triggering any
        relevant events.
        """
        if np.ndim(value) == 0 and len(self.axes) == 1:
            self.position = [self.axes[0].index2value(value)]
        elif len(self.axes) != len(value):
            raise ValueError()
        else:
            p = []
            for i in range(len(self.axes)):
                p.append(self.axes[i].index2value(value[i]))
            self.position = p

    indices = property(lambda s: s._get_indices(),
                       lambda s, v: s._set_indices(v))

    def _pos_changed(self):
        """Call when the position of the widget has changed. It triggers the
        relevant events, and updates the patch position.
        """
        if self._navigating:
            with self.axes_manager.events.indices_changed.suppress_callback(
                    self._on_navigate):
                for i in range(len(self.axes)):
                    self.axes[i].value = self._pos[i]
        self.events.moved.trigger(self)
        self.events.changed.trigger(self)
        self._update_patch_position()

    def _validate_pos(self, pos):
        """Validates the passed position. Depending on the position and the
        implementation, this can either fire a ValueError, or return a modified
        position that has valid values. Or simply return the unmodified
        position if everything is ok.

        This default implementation bounds the position within the axes limits.
        """
        if len(pos) != len(self.axes):
            raise ValueError()
        pos = np.maximum(pos, [ax.low_value for ax in self.axes])
        pos = np.minimum(pos, [ax.high_value for ax in self.axes])
        if self.snap_position:
            pos = self._do_snap_position(pos)
        return pos

    def _get_position(self):
        """Providies the position of the widget (by values) in a tuple.
        """
        return tuple(
            self._pos.tolist())  # Don't pass reference, and make it clear

    def _set_position(self, position):
        """Sets the position of the widget (by values). The dimensions should
        correspond to that of the 'axes' attribute. Calls _pos_changed if the
        value has changed, which is then responsible for triggering any
        relevant events.
        """
        position = self._validate_pos(position)
        if np.any(self._pos != position):
            self._pos = np.array(position)
            self._pos_changed()

    position = property(lambda s: s._get_position(),
                        lambda s, v: s._set_position(v))

    def _do_snap_position(self, value=None):
        """Snaps position to axes grid. Returns snapped value. If value is
        passed as an argument, the internal state is left untouched, if not
        the position attribute is updated to the snapped value.
        """
        value = np.array(value) if value is not None else self._pos
        for i, ax in enumerate(self.axes):
            value[i] = ax.index2value(ax.value2index(value[i]))
        return value

    def _set_snap_position(self, value):
        self._snap_position = value
        if value:
            snap_value = self._do_snap_position(self._pos)
            if np.any(self._pos != snap_value):
                self._pos = snap_value
                self._pos_changed()

    snap_position = property(lambda s: s._snap_position,
                             lambda s, v: s._set_snap_position(v))

    def connect(self, ax):
        super(DraggableWidgetBase, self).connect(ax)
        canvas = ax.figure.canvas
        self.cids.append(
            canvas.mpl_connect('motion_notify_event', self._onmousemove))
        self.cids.append(canvas.mpl_connect('pick_event', self.onpick))
        self.cids.append(canvas.mpl_connect(
            'button_release_event', self.button_release))

    def _on_navigate(self, axes_manager):
        if axes_manager is self.axes_manager:
            p = self._pos.tolist()
            for i, a in enumerate(self.axes):
                p[i] = a.value
            self.position = p    # Use property to trigger events

    def onpick(self, event):
        # Callback for MPL pick event
        self.picked = (event.artist in self.patch)
        self._selected_artist = event.artist
        if hasattr(super(DraggableWidgetBase, self), 'onpick'):
            super(DraggableWidgetBase, self).onpick(event)
        self.selected = self.picked

    def _onmousemove(self, event):
        """Callback for mouse movement. For dragging, the implementor would
        normally check that the widget is picked, and that the event.inaxes
        Axes equals self.ax.
        """
        # This method must be provided by the subclass
        pass

    def _update_patch_position(self):
        """Updates the position of the patch on the plot.
        """
        # This method must be provided by the subclass
        pass

    def _update_patch_geometry(self):
        """Updates all geometry of the patch on the plot.
        """
        self._update_patch_position()

    def button_release(self, event):
        """whenever a mouse button is released"""
        if event.button != 1:
            return
        if self.picked is True:
            self.picked = False


class Widget1DBase(DraggableWidgetBase):

    """A base class for 1D widgets.

    It sets the right dimensions for size and
    position, adds the 'border_thickness' attribute and initalizes the 'axes'
    attribute to the first two navigation axes if possible, if not, the two
    first signal_axes are used. Other than that it mainly supplies common
    utility functions for inheritors, and implements required functions for
    ResizableDraggableWidgetBase.

    The implementation for ResizableDraggableWidgetBase methods all assume that
    a Rectangle patch will be used, centered on position. If not, the
    inheriting class will have to override those as applicable.
    """

    def _set_position(self, position):
        try:
            len(position)
        except TypeError:
            position = (position,)
        super(Widget1DBase, self)._set_position(position)

    def _validate_pos(self, pos):
        pos = np.maximum(pos, self.axes[0].low_value)
        pos = np.minimum(pos, self.axes[0].high_value)
        return super(Widget1DBase, self)._validate_pos(pos)


class ResizableDraggableWidgetBase(DraggableWidgetBase):

    """Adds the `size` property and get_size_in_axes method, and adds a
    framework for letting the user resize the patch, including resizing by
    key strokes ('+', '-'). Also adds the 'resized' event.

    Utility functions for resizing are implemented by `increase_size` and
    `decrease_size`, which will in-/decrement the size by 1. Other utility
    functions include `get_centre` and `get_centre_indices` which returns the
    center position, and the internal _apply_changes which helps make sure that
    only one 'changed' event is fired for a combined move and resize.

    Any inheritors must override these methods:
        _update_patch_position(self)
        _update_patch_size(self)
        _update_patch_geometry(self)
        _set_patch(self)
    """

    def __init__(self, axes_manager, **kwargs):
        super(ResizableDraggableWidgetBase, self).__init__(
            axes_manager, **kwargs)
        if not self.axes:
            self._size = np.array([1])
        self.size_step = 1      # = one step in index space
        self._snap_size = True
        self.events.resized = Event(doc="""
            Event that triggers when the widget was resized.

            The event triggers after the internal state of the widget has been
            updated. This event does not differentiate on how the size of
            the widget was changed, so it is the responsibility of the user
            to suppress events as neccessary to avoid closed loops etc.

            Arguments:
            ----------
                obj:
                    The widget that was resized.
            """, arguments=['obj'])
        self.no_events_while_dragging = False
        self._drag_store = None

    def _set_axes(self, axes):
        super(ResizableDraggableWidgetBase, self)._set_axes(axes)
        if self.axes:
            self._size = np.array([ax.scale for ax in self.axes])

    def _get_size(self):
        """Getter for 'size' property. Returns the size as a tuple (to prevent
        unintended in-place changes).
        """
        return tuple(self._size.tolist())

    def _set_size(self, value):
        """Setter for the 'size' property.

        Calls _size_changed to handle size change, if the value has changed.

        """
        value = np.minimum(value, [ax.size * ax.scale for ax in self.axes])
        value = np.maximum(value,
                           self.size_step * [ax.scale for ax in self.axes])
        if self.snap_size:
            value = self._do_snap_size(value)
        if np.any(self._size != value):
            self._size = value
            self._size_changed()

    size = property(lambda s: s._get_size(), lambda s, v: s._set_size(v))

    def _do_snap_size(self, value=None):
        value = np.array(value) if value is not None else self._size
        for i, ax in enumerate(self.axes):
            value[i] = round(value[i] / ax.scale) * ax.scale
        return value

    def _set_snap_size(self, value):
        self._snap_size = value
        if value:
            snap_value = self._do_snap_size(self._size)
            if np.any(self._size != snap_value):
                self._size = snap_value
                self._size_changed()

    snap_size = property(lambda s: s._snap_size,
                         lambda s, v: s._set_snap_size(v))

    def _set_snap_all(self, value):
        # Snap position first, as snapped size can depend on position.
        self.snap_position = value
        self.snap_size = value

    snap_all = property(lambda s: s.snap_size and s.snap_position,
                        lambda s, v: s._set_snap_all(v))

    def increase_size(self):
        """Increment all sizes by 1. Applied via 'size' property.
        """
        self.size = np.array(self.size) + \
            self.size_step * np.array([a.scale for a in self.axes])

    def decrease_size(self):
        """Decrement all sizes by 1. Applied via 'size' property.
        """
        self.size = np.array(self.size) - \
            self.size_step * np.array([a.scale for a in self.axes])

    def _size_changed(self):
        """Triggers resize and changed events, and updates the patch.
        """
        self.events.resized.trigger(self)
        self.events.changed.trigger(self)
        self._update_patch_size()

    def get_size_in_indices(self):
        """Gets the size property converted to the index space (via 'axes'
        attribute).
        """
        s = list()
        for i in range(len(self.axes)):
            s.append(int(round(self._size[i] / self.axes[i].scale)))
        return np.array(s)

    def set_size_in_indices(self, value):
        """Sets the size property converted to the index space (via 'axes'
        attribute).
        """
        s = list()
        for i in range(len(self.axes)):
            s.append(int(round(value[i] * self.axes[i].scale)))
        self.size = s   # Use property to get full processing

    def get_centre(self):
        """Get's the center indices. The default implementation is simply the
        position + half the size in axes space, which should work for any
        symmetric widget, but more advanced widgets will need to decide whether
        to return the center of gravity or the geometrical center of the
        bounds.
        """
        return self._pos + self._size() / 2.0

    def get_centre_index(self):
        """Get's the center position (in index space). The default
        implementation is simply the indices + half the size, which should
        work for any symmetric widget, but more advanced widgets will need to
        decide whether to return the center of gravity or the geometrical
        center of the bounds.
        """
        return self.indices + self.get_size_in_indices() / 2.0

    def _update_patch_size(self):
        """Updates the size of the patch on the plot.
        """
        # This method must be provided by the subclass
        pass

    def _update_patch_geometry(self):
        """Updates all geometry of the patch on the plot.
        """
        # This method must be provided by the subclass
        pass

    def on_key_press(self, event):
        if event.key == "+":
            self.increase_size()
        if event.key == "-":
            self.decrease_size()

    def connect(self, ax):
        super(ResizableDraggableWidgetBase, self).connect(ax)
        canvas = ax.figure.canvas
        self.cids.append(canvas.mpl_connect('key_press_event',
                                            self.on_key_press))

    def onpick(self, event):
        if hasattr(super(ResizableDraggableWidgetBase, self), 'onpick'):
            super(ResizableDraggableWidgetBase, self).onpick(event)
        if self.picked:
            self._drag_store = (self.position, self.size)

    def _apply_changes(self, old_size, old_position):
        """Evalutes whether the widget has been moved/resized, and triggers
        the correct events and updates the patch geometry. This function has
        the advantage that the geometry is updated only once, preventing
        flickering, and the 'changed' event only fires once.
        """
        moved = self.position != old_position
        resized = self.size != old_size
        if moved:
            if self._navigating:
                e = self.axes_manager.events.indices_changed
                with e.suppress_callback(self._on_navigate):
                    for i in range(len(self.axes)):
                        self.axes[i].index = self.indices[i]
        if moved or resized:
            # Update patch first
            if moved and resized:
                self._update_patch_geometry()
            elif moved:
                self._update_patch_position()
            else:
                self._update_patch_size()
            # Then fire events
            if not self.no_events_while_dragging or not self.picked:
                if moved:
                    self.events.moved.trigger(self)
                if resized:
                    self.events.resized.trigger(self)
                self.events.changed.trigger(self)

    def button_release(self, event):
        """whenever a mouse button is released"""
        picked = self.picked
        super(ResizableDraggableWidgetBase, self).button_release(event)
        if event.button != 1:
            return
        if picked and self.picked is False:
            if self.no_events_while_dragging and self._drag_store:
                self._apply_changes(*self._drag_store)


class Widget2DBase(ResizableDraggableWidgetBase):

    """A base class for 2D widgets. It sets the right dimensions for size and
    position, adds the 'border_thickness' attribute and initalizes the 'axes'
    attribute to the first two navigation axes if possible, if not, the two
    first signal_axes are used. Other than that it mainly supplies common
    utility functions for inheritors, and implements required functions for
    ResizableDraggableWidgetBase.

    The implementation for ResizableDraggableWidgetBase methods all assume that
    a Rectangle patch will be used, centered on position. If not, the
    inheriting class will have to override those as applicable.
    """

    def __init__(self, axes_manager, **kwargs):
        super(Widget2DBase, self).__init__(axes_manager, **kwargs)
        self.border_thickness = 2

        # Set default axes
        if self.axes_manager is not None:
            if self.axes_manager.navigation_dimension > 1:
                self.axes = self.axes_manager.navigation_axes[0:2]
            elif self.axes_manager.signal_dimension > 1:
                self.axes = self.axes_manager.signal_axes[0:2]
            elif len(self.axes_manager.shape) > 1:
                self.axes = (self.axes_manager.signal_axes +
                             self.axes_manager.navigation_axes)
            else:
                raise ValueError("2D widget needs at least two axes!")
        else:
            self._pos = np.array([0, 0])
            self._size = np.array([1, 1])

    def _get_patch_xy(self):
        """Returns the xy position of the widget. In this default
        implementation, the widget is centered on the position.
        """
        return self._pos - self._size / 2.

    def _get_patch_bounds(self):
        """Returns the bounds of the patch in the form of a tuple in the order
        left, top, width, height. In matplotlib, 'bottom' is used instead of
        'top' as the naming assumes an upwards pointing y-axis, meaning the
        lowest value corresponds to bottom. However, our widgets will normally
        only go on images (which has an inverted y-axis in MPL by default), so
        we define the lowest value to be termed 'top'.
        """
        xy = self._get_patch_xy()
        xs, ys = self.size
        return (xy[0], xy[1], xs, ys)        # x,y,w,h

    def _update_patch_position(self):
        if self.is_on() and self.patch:
            self.patch[0].set_xy(self._get_patch_xy())
            self.draw_patch()

    def _update_patch_size(self):
        self._update_patch_geometry()

    def _update_patch_geometry(self):
        if self.is_on() and self.patch:
            self.patch[0].set_bounds(*self._get_patch_bounds())
            self.draw_patch()


class ResizersMixin(object):
    """
    Widget mix-in for adding resizing manipulation handles.

    The default handles are green boxes displayed on the outside corners of the
    boundaries. By default, the handles are only displayed when the widget is
    selected (`picked` in matplotlib terminology).

    Attributes
    ----------
    resizers : bool
        Property that determines whether the resizer handles should be used
    resize_color : matplotlib color
        The color of the resize handles.
    resize_pixel_size : tuple or None
        Size of the resize handles in screen pixels. If None, it is set
        equal to the size of one 'data-pixel' (image pixel size).
    resizer_picked : False or int
        Inidcates which, if any, resizer was selected the last time the
        widget was picked. `False` if another patch was picked, or the
        index of the resizer handle that was picked.

    """

    def __init__(self, resizers=True, **kwargs):
        super(ResizersMixin, self).__init__(**kwargs)
        self.resizer_picked = False
        self.pick_offset = (0, 0)
        self.resize_color = 'lime'
        self.resize_pixel_size = (5, 5)  # Set to None to make one data pixel
        self._resizers = resizers
        self._resizer_handles = []
        self._resizers_on = False
        # The `_resizers_on` attribute reflects whether handles are actually on
        # as compared to `_resizers` which is whether the user wants them on.
        # The difference is e.g. for turning on and off handles when the
        # widget is selected/deselected.

    @property
    def resizers(self):
        return self._resizers

    @resizers.setter
    def resizers(self, value):
        if self._resizers != value:
            self._resizers = value
            self._set_resizers(value, self.ax)

    def _update_resizers(self):
        """Update resizer handles' patch geometry.
        """
        pos = self._get_resizer_pos()
        rsize = self._get_resizer_size()
        for i, r in enumerate(self._resizer_handles):
            r.set_xy(pos[i])
            r.set_width(rsize[0])
            r.set_height(rsize[1])

    def _set_resizers(self, value, ax):
        """Turns the resizers on/off, in much the same way that _set_patch
        works.
        """
        if ax is not None:
            if value:
                for r in self._resizer_handles:
                    ax.add_artist(r)
                    r.set_animated(self.blit)
            else:
                for container in [
                        ax.patches,
                        ax.lines,
                        ax.artists,
                        ax.texts]:
                    for r in self._resizer_handles:
                        if r in container:
                            container.remove(r)
            self._resizers_on = value
            self.draw_patch()

    def _get_resizer_size(self):
        """Gets the size of the resizer handles in axes coordinates. If
        'resize_pixel_size' is None, a size of one pixel will be used.
        """
        invtrans = self.ax.transData.inverted()
        if self.resize_pixel_size is None:
            rsize = [ax.scale for ax in self.axes]
        else:
            rsize = np.abs(invtrans.transform(self.resize_pixel_size) -
                           invtrans.transform((0, 0)))
        return rsize

    def _get_resizer_offset(self):
        """Utility for getting the distance from the boundary box to the
        center of the resize handles.
        """
        invtrans = self.ax.transData.inverted()
        border = self.border_thickness
        # Transform the border thickness into data values
        dl = np.abs(invtrans.transform((border, border)) -
                    invtrans.transform((0, 0))) / 2
        rsize = self._get_resizer_size()
        return rsize / 2 + dl

    def _get_resizer_pos(self):
        """Get the positions of the resizer handles.
        """
        invtrans = self.ax.transData.inverted()
        border = self.border_thickness
        # Transform the border thickness into data values
        dl = np.abs(invtrans.transform((border, border)) -
                    invtrans.transform((0, 0))) / 2
        rsize = self._get_resizer_size()
        xs, ys = self._size

        positions = []
        rp = np.array(self._get_patch_xy())
        p = rp - rsize + dl                         # Top left
        positions.append(p)
        p = rp + (xs - dl[0], -rsize[1] + dl[1])    # Top right
        positions.append(p)
        p = rp + (-rsize[0] + dl[0], ys - dl[1])    # Bottom left
        positions.append(p)
        p = rp + (xs - dl[0], ys - dl[1])           # Bottom right
        positions.append(p)
        return positions

    def _set_patch(self):
        """Creates the resizer handles, irregardless of whether they will be
        used or not.
        """
        if hasattr(super(ResizersMixin, self), '_set_patch'):
            super(ResizersMixin, self)._set_patch()

        if self._resizer_handles:
            self._set_resizers(False, self.ax)
        self._resizer_handles = []
        rsize = self._get_resizer_size()
        pos = self._get_resizer_pos()
        for i in range(len(pos)):
            r = plt.Rectangle(pos[i], rsize[0], rsize[1], animated=self.blit,
                              fill=True, lw=0, fc=self.resize_color,
                              picker=True,)
            self._resizer_handles.append(r)

    def set_on(self, value):
        """Turns on/off resizers whet widget is turned on/off.
        """
        if self.resizers and value != self._resizers_on:
            self._set_resizers(value, self.ax)
        if hasattr(super(ResizersMixin, self), 'set_on'):
            super(ResizersMixin, self).set_on(value)

    def onpick(self, event):
        """Picking of main patch is same as for widget base, but this also
        handles picking of the resize handles. If a resize handles is picked,
        `picked` is set to `True`, and `resizer_picked` is set to an integer
        indicating which handle was picked (0-3 for top left, top right, bottom
        left, bottom right). It is set to `False` if another widget was picked.

        If the main patch is picked, the offset from the picked pixel to the
        `position` is stored in `pick_offset`. This can be used in e.g.
        `_onmousemove` to ease dragging code (prevent widget center/corner
        snapping to mouse).
        """
        if event.artist in self._resizer_handles:
            corner = self._resizer_handles.index(event.artist)
            self.resizer_picked = corner
            self.picked = True
        elif self.picked:
            if self.resizers and not self._resizers_on:
                self._set_resizers(True, self.ax)
            x = event.mouseevent.xdata
            y = event.mouseevent.ydata
            self.pick_offset = (x - self._pos[0], y - self._pos[1])
            self.resizer_picked = False
        else:
            self._set_resizers(False, self.ax)
        if hasattr(super(ResizersMixin, self), 'onpick'):
            super(ResizersMixin, self).onpick(event)

    def _add_patch_to(self, ax):
        """Same as widget base, but also adds resizers if 'resizers' property
        is True.
        """
        if self.resizers:
            self._set_resizers(True, ax)
        if hasattr(super(ResizersMixin, self), '_add_patch_to'):
            super(ResizersMixin, self)._add_patch_to(ax)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy import model
from hyperspy.exceptions import NavigationSizeError


def AIC(model):
    """Calculates the Akaike information criterion

    AIC = 2 k - 2 ln(L)

    where
    L is the maximum likelihood function value,
    k is the number of free parameters.
    """

    # maybe should not have any Offset components?
    # more than a single pixel, needs iterating, don't do that for now
    if model.axes_manager.navigation_size:
        raise NavigationSizeError(model.axes_manager.navigation_size, 0)

    model._set_p0()  # correctly set the parameters (numbers / values)
    lnL = model._poisson_likelihood_function(
        model.p0,
        model.axis.axis[
            model.channel_switches])
    k = len(model.p0) + 1  # +1 for the variance
    return 2 * k - 2 * lnL


def AICc(model):
    _aic = AIC(model)
    n = model.axes_manager.signal_size
    k = len(model.p0) + 1
    return _aic + (2. * k * (k + 1)) / (n - k - 1)


def BIC(model):
    """Calculates the Bayesian information criterion

    BIC = -2 * ln(L) + k * ln(n)

    where
    L is the maximum likelihood function,
    k is the number of free parameters, and
    n is the number of data points (observations) / sample size.
    """
    # maybe should not have any Offset components?
    # more than a single pixel, needs iterating, don't do that for now
    if model.axes_manager.navigation_size:
        raise NavigationSizeError(model.axes_manager.navigation_size, 0)

    model._set_p0()  # correctly set the parameters (numbers / values)
    lnL = model._poisson_likelihood_function(
        model.p0,
        model.axis.axis[
            model.channel_switches])
    n = model.axes_manager.signal_size
    k = len(model.p0) + 1
    return k * np.log(n) - 2. * lnL



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime as dt
import codecs
import os
import logging

import numpy as np
from traits.api import Undefined

from hyperspy import Release
from hyperspy.misc.utils import DictionaryTreeBrowser

_logger = logging.getLogger(__name__)

# Plugin characteristics
# ----------------------
format_name = 'MSA'
description = ''
full_support = False
file_extensions = ('msa', 'ems', 'mas', 'emsa', 'EMS', 'MAS', 'EMSA', 'MSA')
default_extension = 0
# Writing capabilities
writes = [(1, 0), ]
# ----------------------

# For a description of the EMSA/MSA format, incluiding the meaning of the
# following keywords:
# http://www.amc.anl.gov/ANLSoftwareLibrary/02-MMSLib/XEDS/EMMFF/EMMFF.IBM/Emmff.Total

US_MONTHS_D2A = {
    "01" : "JAN",
    "02" : "FEB",
    "03": "MAR",
    "04": "APR",
    "05": "MAY",
    "06": "JUN",
    "07": "JUL",
    "08": "AUG",
    "09": "SEP",
    "10": "OCT",
    "11": "NOV",
    "12": "DEC", }

US_MONTH_A2D = dict([reversed(i) for i in US_MONTHS_D2A.items()])

keywords = {
    # Required parameters
    'FORMAT': {'dtype': str, 'mapped_to': None},
    'VERSION': {'dtype': str, 'mapped_to': None},
    'TITLE': {'dtype': str, 'mapped_to': 'General.title'},
    'DATE': {'dtype': str, 'mapped_to': None},
    'TIME': {'dtype': str, 'mapped_to': None},
    'OWNER': {'dtype': str, 'mapped_to': None},
    'NPOINTS': {'dtype': float, 'mapped_to': None},
    'NCOLUMNS': {'dtype': float, 'mapped_to': None},
    'DATATYPE': {'dtype': str, 'mapped_to': None},
    'XPERCHAN': {'dtype': float, 'mapped_to': None},
    'OFFSET': {'dtype': float, 'mapped_to': None},
    # Optional parameters
    # Signal1D characteristics
    'SIGNALTYPE': {'dtype': str, 'mapped_to':
                   'Signal.signal_type'},
    'XLABEL': {'dtype': str, 'mapped_to': None},
    'YLABEL': {'dtype': str, 'mapped_to': None},
    'XUNITS': {'dtype': str, 'mapped_to': None},
    'YUNITS': {'dtype': str, 'mapped_to': None},
    'CHOFFSET': {'dtype': float, 'mapped_to': None},
    'COMMENT': {'dtype': str, 'mapped_to': None},
    # Microscope
    'BEAMKV': {'dtype': float, 'mapped_to':
               'Acquisition_instrument.TEM.beam_energy'},
    'EMISSION': {'dtype': float, 'mapped_to': None},
    'PROBECUR': {'dtype': float, 'mapped_to':
                 'Acquisition_instrument.TEM.beam_current'},
    'BEAMDIAM': {'dtype': float, 'mapped_to': None},
    'MAGCAM': {'dtype': float, 'mapped_to': None},
    'OPERMODE': {'dtype': str, 'mapped_to': None},
    'CONVANGLE': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.convergence_angle'},

    # Specimen
    'THICKNESS': {'dtype': float, 'mapped_to':
                  'Sample.thickness'},
    'XTILTSTGE': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.Stage.tilt_alpha'},
    'YTILTSTGE': {'dtype': float, 'mapped_to': None},
    'XPOSITION': {'dtype': float, 'mapped_to': None},
    'YPOSITION': {'dtype': float, 'mapped_to': None},
    'ZPOSITION': {'dtype': float, 'mapped_to': None},

    # EELS
    # in ms:
    'INTEGTIME': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.Detector.EELS.exposure'},
    # in ms:
    'DWELLTIME': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.Detector.EELS.dwell_time'},
    'COLLANGLE': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.Detector.EELS.collection_angle'},
    'ELSDET': {'dtype': str, 'mapped_to': None},

    # EDS
    'ELEVANGLE': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.Detector.EDS.elevation_angle'},
    'AZIMANGLE': {'dtype': float, 'mapped_to':
                  'Acquisition_instrument.TEM.Detector.EDS.azimuth_angle'},
    'SOLIDANGLE': {'dtype': float, 'mapped_to':
                   'Acquisition_instrument.TEM.Detector.EDS.solid_angle'},
    'LIVETIME': {'dtype': float, 'mapped_to':
                 'Acquisition_instrument.TEM.Detector.EDS.live_time'},
    'REALTIME': {'dtype': float, 'mapped_to':
                 'Acquisition_instrument.TEM.Detector.EDS.real_time'},
    'FWHMMNKA': {'dtype': float, 'mapped_to':
                 'Acquisition_instrument.TEM.Detector.EDS.' +
                 'energy_resolution_MnKa'},
    'TBEWIND': {'dtype': float, 'mapped_to': None},
    'TAUWIND': {'dtype': float, 'mapped_to': None},
    'TDEADLYR': {'dtype': float, 'mapped_to': None},
    'TACTLYR': {'dtype': float, 'mapped_to': None},
    'TALWIND': {'dtype': float, 'mapped_to': None},
    'TPYWIND': {'dtype': float, 'mapped_to': None},
    'TBNWIND': {'dtype': float, 'mapped_to': None},
    'TDIWIND': {'dtype': float, 'mapped_to': None},
    'THCWIND': {'dtype': float, 'mapped_to': None},
    'EDSDET': {'dtype': str, 'mapped_to':
               'Acquisition_instrument.TEM.Detector.EDS.EDS_det'},
}


def parse_msa_string(string, filename=None):
    """Parse an EMSA/MSA file content.

    Parameters
    ----------
    string: string or file object
        It must complain with the EMSA/MSA standard.
    filename: string or None
        The filename.

    Returns:
    --------
    file_data_list: list
        The list containts a dictionary that contains the parsed
        information. It can be used to create a `:class:BaseSignal`
        using `:func:hyperspy.io.dict2signal`.

    """
    if not hasattr(string, "readlines"):
        string = string.splitlines()
    parameters = {}
    mapped = DictionaryTreeBrowser({})
    y = []
    # Read the keywords
    data_section = False
    for line in string:
        if data_section is False:
            if line[0] == "#":
                try:
                    key, value = line.split(': ')
                    value = value.strip()
                except ValueError:
                    key = line
                    value = None
                key = key.strip('#').strip()

                if key != 'SPECTRUM':
                    parameters[key] = value
                else:
                    data_section = True
        else:
            # Read the data
            if line[0] != "#" and line.strip():
                if parameters['DATATYPE'] == 'XY':
                    xy = line.replace(',', ' ').strip().split()
                    y.append(float(xy[1]))
                elif parameters['DATATYPE'] == 'Y':
                    data = [
                        float(i) for i in line.replace(',', ' ').strip().split()]
                    y.extend(data)
    # We rewrite the format value to be sure that it complies with the
    # standard, because it will be used by the writer routine
    parameters['FORMAT'] = "EMSA/MAS Spectral Data File"

    # Convert the parameters to the right type and map some
    # TODO: the msa format seems to support specifying the units of some
    # parametes. We should add this feature here
    for parameter, value in parameters.items():
        # Some parameters names can contain the units information
        # e.g. #AZIMANGLE-dg: 90.
        if '-' in parameter:
            clean_par, units = parameter.split('-')
            clean_par, units = clean_par.strip(), units.strip()
        else:
            clean_par, units = parameter, None
        if clean_par in keywords:
            try:
                parameters[parameter] = keywords[clean_par]['dtype'](value)
            except BaseException:
                # Normally the offending mispelling is a space in the scientic
                # notation, e.g. 2.0 E-06, so we try to correct for it
                try:
                    parameters[parameter] = keywords[clean_par]['dtype'](
                        value.replace(' ', ''))
                except BaseException:
                    _logger.exception(
                        "The %s keyword value, %s could not be converted to "
                        "the right type", parameter, value)

            if keywords[clean_par]['mapped_to'] is not None:
                mapped.set_item(keywords[clean_par]['mapped_to'],
                                parameters[parameter])
                if units is not None:
                    mapped.set_item(keywords[clean_par]['mapped_to'] +
                                    '_units', units)
    if 'TIME' in parameters and parameters['TIME']:
        try:
            time = dt.strptime(parameters['TIME'], "%H:%M")
            mapped.set_item('General.time', time.time().isoformat())
        except ValueError as e:
            _logger.warning('Possible malformed TIME field in msa file. The time information could not be retrieved.: %s' % e)
    else:
        _logger.warning('TIME information missing.')

    malformed_date_error = 'Possibly malformed DATE in msa file. The date information could not be retrieved.'
    if "DATE" in parameters and parameters["DATE"]:
        try:
            day, month, year = parameters["DATE"].split("-")
            if month.upper() in US_MONTH_A2D:
                month = US_MONTH_A2D[month.upper()]
                date = dt.strptime("-".join((day, month, year)), "%d-%m-%Y")
                mapped.set_item('General.date', date.date().isoformat())
            else:
                    _logger.warning(malformed_date_error)
        except ValueError as e: # Error raised if split does not return 3 elements in this case
            _logger.warning(malformed_date_error + ": %s" % e)


    axes = [{
        'size': len(y),
        'index_in_array': 0,
        'name': parameters['XLABEL'] if 'XLABEL' in parameters else '',
        'scale': parameters['XPERCHAN'] if 'XPERCHAN' in parameters else 1,
        'offset': parameters['OFFSET'] if 'OFFSET' in parameters else 0,
        'units': parameters['XUNITS'] if 'XUNITS' in parameters else '',
    }]
    if filename is not None:
        mapped.set_item('General.original_filename',
                        os.path.split(filename)[1])
    mapped.set_item('Signal.record_by', 'spectrum')
    if mapped.has_item('Signal.signal_type'):
        if mapped.Signal.signal_type == 'ELS':
            mapped.Signal.signal_type = 'EELS'
        if mapped.Signal.signal_type in ['EDX', 'XEDS']:
            mapped.Signal.signal_type = 'EDS'
    else:
        # Defaulting to EELS looks reasonable
        mapped.set_item('Signal.signal_type', 'EELS')
    if 'YUNITS' in parameters.keys():
        yunits = "(%s)" % parameters['YUNITS']
    else:
        yunits = ""
    if 'YLABEL' in parameters.keys():
        quantity = "%s" % parameters['YLABEL']
    else:
        if mapped.Signal.signal_type == 'EELS':
            quantity = 'Electrons'
            if not yunits:
                yunits = "(Counts)"
        elif 'EDS' in mapped.Signal.signal_type:
            quantity = 'X-rays'
            if not yunits:
                yunits = "(Counts)"
        else:
            quantity = ""
    if quantity or yunits:
        quantity_units = "%s %s" % (quantity, yunits)
        mapped.set_item('Signal.quantity', quantity_units.strip())

    dictionary = {
        'data': np.array(y),
        'axes': axes,
        'metadata': mapped.as_dictionary(),
        'original_metadata': parameters
    }
    file_data_list = [dictionary, ]
    return file_data_list


def file_reader(filename, encoding='latin-1', **kwds):
    with codecs.open(
            filename,
            encoding=encoding,
            errors='replace') as spectrum_file:
        return parse_msa_string(string=spectrum_file,
                                filename=filename)


def file_writer(filename, signal, format=None, separator=', ',
                encoding='latin-1'):
    loc_kwds = {}
    FORMAT = "EMSA/MAS Spectral Data File"
    md = signal.metadata
    if hasattr(signal.original_metadata, 'FORMAT') and \
            signal.original_metadata.FORMAT == FORMAT:
        loc_kwds = signal.original_metadata.as_dictionary()
        if format is not None:
            loc_kwds['DATATYPE'] = format
        else:
            if 'DATATYPE' in loc_kwds:
                format = loc_kwds['DATATYPE']
    else:
        if format is None:
            format = 'Y'
        if md.has_item("General.date"):
            date = dt.strptime(md.General.date, "%Y-%m-%d")
            date_str = date.strftime("%d-%m-%Y")
            day, month, year = date_str.split("-")
            month = US_MONTHS_D2A[month]
            loc_kwds['DATE'] = "-".join((day, month, year)) 
        if md.has_item("General.time"):
            time = dt.strptime(md.General.time, "%H:%M:%S")
            loc_kwds['TIME'] = time.strftime("%H:%M")
    keys_from_signal = {
        # Required parameters
        'FORMAT': FORMAT,
        'VERSION': '1.0',
        # 'TITLE' : signal.title[:64] if hasattr(signal, "title") else '',
        'DATE': '',
        'TIME': '',
        'OWNER': '',
        'NPOINTS': signal.axes_manager._axes[0].size,
        'NCOLUMNS': 1,
        'DATATYPE': format,
        'SIGNALTYPE': signal.metadata.Signal.signal_type,
        'XPERCHAN': signal.axes_manager._axes[0].scale,
        'OFFSET': signal.axes_manager._axes[0].offset,
        # Signal1D characteristics

        'XLABEL': signal.axes_manager._axes[0].name
        if signal.axes_manager._axes[0].name is not Undefined
        else "",

        #        'YLABEL' : '',
        'XUNITS': signal.axes_manager._axes[0].units
        if signal.axes_manager._axes[0].units is not Undefined
        else "",
        #        'YUNITS' : '',
        'COMMENT': 'File created by HyperSpy version %s' % Release.version,
        # Microscope
        #        'BEAMKV' : ,
        #        'EMISSION' : ,
        #        'PROBECUR' : ,
        #        'BEAMDIAM' : ,
        #        'MAGCAM' : ,
        #        'OPERMODE' : ,
        #        'CONVANGLE' : ,
        # Specimen
        #        'THICKNESS' : ,
        #        'XTILTSTGE' : ,
        #        'YTILTSTGE' : ,
        #        'XPOSITION' : ,
        #        'YPOSITION' : ,
        #        'ZPOSITION' : ,
        #
        # EELS
        # 'INTEGTIME' : , # in ms
        # 'DWELLTIME' : , # in ms
        #        'COLLANGLE' : ,
        #        'ELSDET' :  ,
    }

    # Update the loc_kwds with the information retrieved from the signal class
    for key, value in keys_from_signal.items():
        if key not in loc_kwds or value != '':
            loc_kwds[key] = value

    for key, dic in keywords.items():

        if dic['mapped_to'] is not None:
            if 'SEM' in signal.metadata.Signal.signal_type:
                dic['mapped_to'] = dic['mapped_to'].replace('TEM', 'SEM')
            if signal.metadata.has_item(dic['mapped_to']):
                loc_kwds[key] = eval('signal.metadata.%s' %
                                     dic['mapped_to'])

    with codecs.open(
            filename,
            'w',
            encoding=encoding,
            errors='ignore') as f:
        # Remove the following keys from loc_kwds if they are in
        # (although they shouldn't)
        for key in ['SPECTRUM', 'ENDOFDATA']:
            if key in loc_kwds:
                del(loc_kwds[key])

        f.write('#%-12s: %s\u000D\u000A' % ('FORMAT', loc_kwds.pop('FORMAT')))
        f.write(
            '#%-12s: %s\u000D\u000A' %
            ('VERSION', loc_kwds.pop('VERSION')))
        for keyword, value in loc_kwds.items():
            f.write('#%-12s: %s\u000D\u000A' % (keyword, value))

        f.write('#%-12s: Spectral Data Starts Here\u000D\u000A' % 'SPECTRUM')

        if format == 'XY':
            for x, y in zip(signal.axes_manager._axes[0].axis, signal.data):
                f.write("%g%s%g" % (x, separator, y))
                f.write('\u000D\u000A')
        elif format == 'Y':
            for y in signal.data:
                f.write('%f%s' % (y, separator))
                f.write('\u000D\u000A')
        else:
            raise ValueError('format must be one of: None, \'XY\' or \'Y\'')

        f.write('#%-12s: End Of Data and File' % 'ENDOFDATA')



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial
import logging

from traits.api import Undefined

from hyperspy.drawing import widgets, signal1d, image
from hyperspy.defaults_parser import preferences


_logger = logging.getLogger(__name__)


class MPL_HyperExplorer(object):

    """

    """

    def __init__(self):
        self.signal_data_function = None
        self.navigator_data_function = None
        # args to pass to `__call__`
        self.signal_data_function_kwargs = {}
        self.axes_manager = None
        self.signal_title = ''
        self.navigator_title = ''
        self.quantity_label = ''
        self.signal_plot = None
        self.navigator_plot = None
        self.axis = None
        self.pointer = None
        self._pointer_nav_dim = None

    def plot_signal(self, **kwargs):
        # This method should be implemented by the subclasses.
        # Doing nothing is good enough for signal_dimension==0 though.
        if self.axes_manager.signal_dimension == 0:
            return
        if self.signal_data_function_kwargs.get('fft_shift', False):
            self.axes_manager = self.axes_manager.deepcopy()
            for axis in self.axes_manager.signal_axes:
                axis.offset = -axis.high_value / 2.

    def plot_navigator(self, title=None, **kwargs):
        """
        Parameters
        ----------
        title : str, optional
            Title of the navigator. The default is None.
        **kwargs : dict
            The kwargs are passed to plot method of
            :py:meth:`hyperspy.drawing.image.ImagePlot` or
            :py:meth:`hyperspy.drawing.signal1d.Signal1DLine`.

        """
        if self.axes_manager.navigation_dimension == 0:
            return
        if self.navigator_data_function is None:
            return
        if self.navigator_data_function == "slider":
            self._get_navigation_sliders()
            return
        title = title or self.signal_title + " Navigator" if self.signal_title else ""

        if len(self.navigator_data_function().shape) == 1:
            # Create the figure
            sf = signal1d.Signal1DFigure(title=title)
            axis = self.axes_manager.navigation_axes[0]
            sf.xlabel = '%s' % str(axis)
            if axis.units is not Undefined:
                sf.xlabel += ' (%s)' % axis.units
            sf.ylabel = r'$\Sigma\mathrm{data\,over\,all\,other\,axes}$'
            sf.axis = axis
            sf.axes_manager = self.axes_manager
            self.navigator_plot = sf

            # Create a line to the left axis with the default indices
            sl = signal1d.Signal1DLine()
            sl.data_function = self.navigator_data_function

            # Set all kwargs value to the image figure before passing the rest
            # of the kwargs to plot method of the image figure
            for key in list(kwargs.keys()):
                if hasattr(sl, key):
                    setattr(sl, key, kwargs.pop(key))
            sl.set_line_properties(color='blue',
                                   type='step')

            # Add the line to the figure
            sf.add_line(sl)
            sf.plot()
            self.pointer.set_mpl_ax(sf.ax)
            if self.axes_manager.navigation_dimension > 1:
                self._get_navigation_sliders()
                for axis in self.axes_manager.navigation_axes[:-2]:
                    axis.events.index_changed.connect(sf.update, [])
                    sf.events.closed.connect(
                        partial(axis.events.index_changed.disconnect,
                                sf.update), [])
            self.navigator_plot = sf
        elif len(self.navigator_data_function().shape) >= 2:
            # Create the figure
            imf = image.ImagePlot(title=title)
            imf.data_function = self.navigator_data_function

            # Set all kwargs value to the image figure before passing the rest
            # of the kwargs to plot method of the image figure
            for key, value in list(kwargs.items()):
                if hasattr(imf, key):
                    setattr(imf, key, kwargs.pop(key))

            # Navigator labels
            if self.axes_manager.navigation_dimension == 1:
                imf.yaxis = self.axes_manager.navigation_axes[0]
                imf.xaxis = self.axes_manager.signal_axes[0]
            elif self.axes_manager.navigation_dimension >= 2:
                imf.yaxis = self.axes_manager.navigation_axes[1]
                imf.xaxis = self.axes_manager.navigation_axes[0]
                if self.axes_manager.navigation_dimension > 2:
                    self._get_navigation_sliders()
                    for axis in self.axes_manager.navigation_axes[2:]:
                        axis.events.index_changed.connect(imf.update, [])
                        imf.events.closed.connect(
                            partial(axis.events.index_changed.disconnect,
                                    imf.update), [])

            if "cmap" not in kwargs.keys() or kwargs['cmap'] is None:
                kwargs["cmap"] = preferences.Plot.cmap_navigator
            imf.plot(**kwargs)
            self.pointer.set_mpl_ax(imf.ax)
            self.navigator_plot = imf

        if self.navigator_plot is not None:
            self.navigator_plot.events.closed.connect(
                self._on_navigator_plot_closing, [])

    def _get_navigation_sliders(self):
        try:
            self.axes_manager.gui_navigation_sliders(
                title=self.signal_title + " navigation sliders")
        except (ValueError, ImportError) as e:
            _logger.warning("Navigation sliders not available. " + str(e))

    def close_navigator_plot(self):
        if self.navigator_plot:
            self.navigator_plot.close()

    @property
    def is_active(self):
        """A plot is active when it has the figure open meaning that it has
        either one of 'signal_plot' or 'navigation_plot' is not None and it
        has a attribute 'figure' which is not None.
        """
        if self.signal_plot and self.signal_plot.figure:
            return True
        elif self.navigator_plot and self.navigator_plot.figure:
            return True
        else:
            return False

    def plot(self, **kwargs):
        # Parse the kwargs for plotting complex data
        for key in ['power_spectrum', 'fft_shift']:
            if key in kwargs:
                self.signal_data_function_kwargs[key] = kwargs.pop(key)
        if self.pointer is None:
            pointer = self.assign_pointer()
            if pointer is not None:
                self.pointer = pointer(self.axes_manager)
                self.pointer.color = 'red'
                self.pointer.connect_navigate()
            self.plot_navigator(**kwargs.pop('navigator_kwds', {}))
            if pointer is not None:
                self.navigator_plot.events.closed.connect(
                    self.pointer.disconnect, [])
        self.plot_signal(**kwargs)

    def assign_pointer(self):
        if self.navigator_data_function is None:
            nav_dim = 0
        elif self.navigator_data_function == "slider":
            nav_dim = 0
        else:
            nav_dim = len(self.navigator_data_function().shape)

        if nav_dim == 2:  # It is an image
            if self.axes_manager.navigation_dimension > 1:
                Pointer = widgets.SquareWidget
            else:  # It is the image of a "spectrum stack"
                Pointer = widgets.HorizontalLineWidget
        elif nav_dim == 1:  # It is a spectrum
            Pointer = widgets.VerticalLineWidget
        else:
            Pointer = None
        self._pointer_nav_dim = nav_dim
        return Pointer

    def _on_navigator_plot_closing(self):
        self.navigator_plot = None

    def _on_signal_plot_closing(self):
        self.signal_plot = None

    def close(self):
        """When closing, we make sure:
        - close the matplotlib figure
        - drawing events are disconnected
        - the attribute 'signal_plot' and 'navigation_plot' are set to None
        """
        if self.is_active:
            if self.signal_plot:
                self.signal_plot.close()
            self.close_navigator_plot()



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.fftpack import fft2
import logging

_logger = logging.getLogger(__name__)


def calculate_carrier_frequency(holo_data, sb_position, scale):
    """
    Calculates fringe carrier frequency of a hologram

    Parameters
    ----------
    holo_data: ndarray
        The data of the hologram.
    sb_position: tuple
        Position of the sideband with the reference to non-shifted FFT
    scale: tuple
        Scale of the axes that will be used for the calculation.

    Returns
    -------
    Carrier frequency
    """

    shape = holo_data.shape
    origins = [np.array((0, 0)),
               np.array((0, shape[1])),
               np.array((shape[0], shape[1])),
               np.array((shape[0], 0))]
    origin_index = np.argmin(
        [np.linalg.norm(origin - sb_position) for origin in origins])
    return np.linalg.norm(np.multiply(
        origins[origin_index] - sb_position, scale))


def estimate_fringe_contrast_fourier(
        holo_data, sb_position, apodization='hanning'):
    """
    Estimates average fringe contrast of a hologram  by dividing amplitude
    of maximum pixel of sideband by amplitude of FFT's origin.

    Parameters
    ----------
    holo_data: ndarray
        The data of the hologram.
    sb_position: tuple
        Position of the sideband with the reference to non-shifted FFT
    apodization: string, None
        Use 'hanning', 'hamming' or None to apply apodization window in real space before FFT
        Apodization is typically needed to suppress the striking  due to sharp edges
        of the which often results in underestimation of the fringe contrast. (Default: 'hanning')

    Returns
    -------
    Fringe contrast as a float
    """

    holo_shape = holo_data.shape

    if apodization:
        if apodization == 'hanning':
            window_x = np.hanning(holo_shape[0])
            window_y = np.hanning(holo_shape[1])
        elif apodization == 'hamming':
            window_x = np.hamming(holo_shape[0])
            window_y = np.hamming(holo_shape[1])
        window_2d = np.sqrt(np.outer(window_x, window_y))
        data = holo_data * window_2d
    else:
        data = holo_data

    fft_exp = fft2(data)

    return 2 * np.abs(fft_exp[tuple(sb_position)]) / np.abs(fft_exp[0, 0])



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import logging

from hyperspy._components.expression import Expression


_logger = logging.getLogger(__name__)


class PowerLaw(Expression):

    r"""Power law component.

    .. math::

        f(x) = A\cdot(x-x_0)^{-r}

    ============= =============
     Variable      Parameter
    ============= =============
     :math:`A`     A
     :math:`r`     r
     :math:`x_0`   origin
    ============= =============


    Parameters
    ----------
    A : float
        Height parameter.
    r : float
        Power law coefficient.
    origin : float
        Location parameter.
    **kwargs
        Extra keyword arguments are passed to the ``Expression`` component.


    The `left_cutoff` parameter can be used to set a lower threshold from which
    the component will return 0.
    """

    def __init__(self, A=10e5, r=3., origin=0., left_cutoff=0.0,
                 module="numexpr", compute_gradients=False, **kwargs):
        super().__init__(
            expression="where(left_cutoff<x, A*(-origin + x)**-r, 0)",
            name="PowerLaw",
            A=A,
            r=r,
            origin=origin,
            left_cutoff=left_cutoff,
            position="origin",
            module=module,
            autodoc=False,
            compute_gradients=compute_gradients,
            **kwargs,
        )

        self.origin.free = False
        self.left_cutoff.free = False

        # Boundaries
        self.A.bmin = 0.
        self.A.bmax = None
        self.r.bmin = 1.
        self.r.bmax = 5.

        self.isbackground = True
        self.convolved = False

    def estimate_parameters(self, signal, x1, x2, only_current=False,
                            out=False):
        """Estimate the parameters for the power law component by the two area
        method.

        Parameters
        ----------
        signal : Signal1D instance
        x1 : float
            Defines the left limit of the spectral range to use for the
            estimation.
        x2 : float
            Defines the right limit of the spectral range to use for the
            estimation.
        only_current : bool
            If False, estimates the parameters for the full dataset.
        out : bool
            If True, returns the result arrays directly without storing in the
            parameter maps/values. The returned order is (A, r).

        Returns
        -------
        {bool, tuple of values}

        """
        super(PowerLaw, self)._estimate_parameters(signal)
        axis = signal.axes_manager.signal_axes[0]
        i1, i2 = axis.value_range_to_indices(x1, x2)
        if not (i2 + i1) % 2 == 0:
            i2 -= 1
        if i2 == i1:
            i2 += 2
        i3 = (i2 + i1) // 2
        x1 = axis.index2value(i1)
        x2 = axis.index2value(i2)
        x3 = axis.index2value(i3)
        if only_current is True:
            s = signal.get_current_signal()
        else:
            s = signal
        if s._lazy:
            import dask.array as da
            log = da.log
            I1 = s.isig[i1:i3].integrate1D(2j).data
            I2 = s.isig[i3:i2].integrate1D(2j).data
        else:
            from hyperspy.signal import BaseSignal
            shape = s.data.shape[:-1]
            I1_s = BaseSignal(np.empty(shape, dtype='float'))
            I2_s = BaseSignal(np.empty(shape, dtype='float'))
            # Use the `out` parameters to avoid doing the deepcopy
            s.isig[i1:i3].integrate1D(2j, out=I1_s)
            s.isig[i3:i2].integrate1D(2j, out=I2_s)
            I1 = I1_s.data
            I2 = I2_s.data
            log = np.log
        with np.errstate(divide='raise'):
            try:
                r = 2 * (log(I1) - log(I2)) / (log(x2) - log(x1))
                k = 1 - r
                A = k * I2 / (x2 ** k - x3 ** k)
                if s._lazy:
                    r = r.map_blocks(np.nan_to_num)
                    A = A.map_blocks(np.nan_to_num)
                else:
                    r = np.nan_to_num(r)
                    A = np.nan_to_num(A)
            except (RuntimeWarning, FloatingPointError):
                _logger.warning('Power-law parameter estimation failed '
                                'because of a "divide-by-zero" error.')
                return False
        if only_current is True:
            self.r.value = r
            self.A.value = A
            return True
        if out:
            return A, r
        else:
            if self.A.map is None:
                self._create_arrays()
            self.A.map['values'][:] = A
            self.A.map['is_set'][:] = True
            self.r.map['values'][:] = r
            self.r.map['is_set'][:] = True
            self.fetch_stored_values()
            return True

    def grad_A(self, x):
        return self.function(x) / self.A.value

    def grad_r(self, x):
        return np.where(x > self.left_cutoff.value, -self.A.value *
                        np.log(x - self.origin.value) *
                        (x - self.origin.value) ** (-self.r.value), 0)

    def grad_origin(self, x):
        return np.where(x > self.left_cutoff.value, self.r.value *
                        (x - self.origin.value) ** (-self.r.value - 1) *
                        self.A.value, 0)



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.misc.utils import (DictionaryTreeBrowser, check_long_string,
                                 replace_html_symbols)
from hyperspy.signal import BaseSignal


class TestDictionaryBrowser:

    def setup_method(self, method):
        tree = DictionaryTreeBrowser(
            {
                "Node1": {"leaf11": 11,
                          "Node11": {"leaf111": 111},
                          },
                "Node2": {"leaf21": 21,
                          "Node21": {"leaf211": 211},
                          },
            })
        self.tree = tree

    def test_add_dictionary(self):
        self.tree.add_dictionary({
            "Node1": {"leaf12": 12,
                      "Node11": {"leaf111": 222,
                                 "Node111": {"leaf1111": 1111}, },
                      },
            "Node3": {
                "leaf31": 31},
        })
        assert (
            {"Node1": {"leaf11": 11,
                       "leaf12": 12,
                       "Node11": {"leaf111": 222,
                                  "Node111": {
                                      "leaf1111": 1111},
                                  },
                       },
             "Node2": {"leaf21": 21,
                       "Node21": {"leaf211": 211},
                       },
             "Node3": {"leaf31": 31},
             } == self.tree.as_dictionary())

    def test_add_signal_in_dictionary(self):
        tree = self.tree
        s = BaseSignal([1., 2, 3])
        s.axes_manager[0].name = 'x'
        s.axes_manager[0].units = 'ly'
        tree.add_dictionary({"_sig_signal name": s._to_dictionary()})
        assert isinstance(tree.signal_name, BaseSignal)
        np.testing.assert_array_equal(tree.signal_name.data, s.data)
        assert (tree.signal_name.metadata.as_dictionary() ==
                s.metadata.as_dictionary())
        assert (tree.signal_name.axes_manager._get_axes_dicts() ==
                s.axes_manager._get_axes_dicts())

    def test_signal_to_dictionary(self):
        tree = self.tree
        s = BaseSignal([1., 2, 3])
        s.axes_manager[0].name = 'x'
        s.axes_manager[0].units = 'ly'
        tree.set_item('Some name', s)
        d = tree.as_dictionary()
        np.testing.assert_array_equal(d['_sig_Some name']['data'], s.data)
        d['_sig_Some name']['data'] = 0
        assert (
            {
                "Node1": {
                    "leaf11": 11,
                    "Node11": {
                        "leaf111": 111},
                },
                "Node2": {
                    "leaf21": 21,
                    "Node21": {
                        "leaf211": 211},
                },
                "_sig_Some name": {
                    'attributes': {'_lazy': False},
                    'axes': [
                        {
                            'name': 'x',
                            'navigate': False,
                                    'offset': 0.0,
                                    'scale': 1.0,
                                    'size': 3,
                                    'units': 'ly'}],
                    'data': 0,
                    'learning_results': {},
                    'metadata': {
                        'General': {
                            'title': ''},
                        'Signal': {
                            'binned': False,
                            'signal_type': ''},
                        '_HyperSpy': {
                            'Folding': {
                                'original_axes_manager': None,
                                'original_shape': None,
                                'unfolded': False,
                                'signal_unfolded': False}}},
                    'original_metadata': {},
                    'tmp_parameters': {}}} ==
            d)

    def _test_date_time(self, dt_str='now'):
        dt0 = np.datetime64(dt_str)
        data_str, time_str = np.datetime_as_string(dt0).split('T')
        self.tree.add_node("General")
        self.tree.General.date = data_str
        self.tree.General.time = time_str

        dt1 = np.datetime64('%sT%s' % (self.tree.General.date,
                                       self.tree.General.time))

        np.testing.assert_equal(dt0, dt1)
        return dt1

    def test_date_time_now(self):
        # not really a test, more a demo to show how to set and use date and
        # time in the DictionaryBrowser
        self._test_date_time()

    def test_date_time_nanosecond_precision(self):
        # not really a test, more a demo to show how to set and use date and
        # time in the DictionaryBrowser
        dt_str = '2016-08-05T10:13:15.450580'
        self._test_date_time(dt_str)

    def test_has_item(self):
        # Check that it finds all actual items:
        assert self.tree.has_item('Node1')
        assert self.tree.has_item('Node1.leaf11')
        assert self.tree.has_item('Node1.Node11')
        assert self.tree.has_item('Node1.Node11.leaf111')
        assert self.tree.has_item('Node2')
        assert self.tree.has_item('Node2.leaf21')
        assert self.tree.has_item('Node2.Node21')
        assert self.tree.has_item('Node2.Node21.leaf211')

        # Check that it doesn't find non-existant ones
        assert not self.tree.has_item('Node3')
        assert not self.tree.has_item('General')
        assert not self.tree.has_item('Node1.leaf21')
        assert not self.tree.has_item('')
        assert not self.tree.has_item('.')
        assert not self.tree.has_item('..Node1')

    def test_get_item(self):
        # Check that it gets all leaf nodes:
        assert self.tree.get_item('Node1.leaf11') == 11
        assert self.tree.get_item('Node1.Node11.leaf111') == 111
        assert self.tree.get_item('Node2.leaf21') == 21
        assert self.tree.get_item('Node2.Node21.leaf211') == 211

        # Check that it gets all leaf nodes, also with given default:
        assert self.tree.get_item('Node1.leaf11', 44) == 11
        assert self.tree.get_item('Node1.Node11.leaf111', 44) == 111
        assert self.tree.get_item('Node2.leaf21', 44) == 21
        assert self.tree.get_item('Node2.Node21.leaf211', 44) == 211

        # Check that it returns the default value for various incorrect paths:
        assert self.tree.get_item('Node1.leaf33', 44) == 44
        assert self.tree.get_item('Node1.leaf11.leaf111', 44) == 44
        assert self.tree.get_item('Node1.Node31.leaf311', 44) == 44
        assert self.tree.get_item('Node1.Node21.leaf311', 44) == 44
        assert self.tree.get_item('.Node1.Node21.leaf311', 44) == 44

    def test_html(self):
        "Test that the method actually runs"
        # We do not have a way to validate html
        # without relying on more dependencies
        tree = self.tree
        tree['<myhtmltag>'] = "5 < 6"
        tree['<mybrokenhtmltag'] = "<hello>"
        tree['mybrokenhtmltag2>'] = ""
        tree._get_html_print_items()

def test_check_long_string():
    max_len = 20
    value = "Hello everyone this is a long string"
    truth, shortened = check_long_string(value, max_len)
    assert truth == False
    assert shortened == 'Hello everyone this is a long string'

    value = "No! It was not a long string! This is a long string!"
    truth, shortened = check_long_string(value, max_len)
    assert truth == True
    assert shortened == 'No! It was not a lon ... is is a long string!'

def test_replace_html_symbols():
    assert '&lt;&gt;&amp' == replace_html_symbols('<>&')
    assert 'no html symbols' == replace_html_symbols('no html symbols')
    assert '&lt;mix&gt;' == replace_html_symbols('<mix>')

def test_add_key_value():
    key = "<foo>"
    value = ">bar<"

    string = """<ul style="margin: 0px; list-style-position: outside;">
        <li style='margin-left:1em; padding-left: 0.5em'>{} = {}</li></ul>
        """.format(replace_html_symbols(key), replace_html_symbols(value))

    assert string == '<ul style="margin: 0px; list-style-position: outside;">\n        <li style=\'margin-left:1em; padding-left: 0.5em\'>&lt;foo&gt; = &gt;bar&lt;</li></ul>\n        '



# -*- coding: utf-8 -*-
# Copyright 2007-2020 The HyperSpy developers
#
# This file is part of  HyperSpy.
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy.decorators import lazifyTestClass


# Dask does not always work nicely with np.errstate,
# see: https://github.com/dask/dask/issues/3245, so
# filter out divide-by-zero warnings that only appear
# when the test is lazy. When the test is not lazy,
# internal use of np.errstate means the warnings never
# appear in the first place.
@pytest.mark.filterwarnings("ignore:invalid value encountered in subtract:RuntimeWarning")
@pytest.mark.filterwarnings("ignore:divide by zero encountered in log:RuntimeWarning")
@lazifyTestClass
class TestCreateEELSModel:

    def setup_method(self, method):
        s = hs.signals.EELSSpectrum(np.zeros(200))
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        s.add_elements(("B", "C"))
        self.s = s

    def test_create_eelsmodel(self):
        from hyperspy.models.eelsmodel import EELSModel
        assert isinstance(self.s.create_model(), EELSModel)

    def test_create_eelsmodel_no_md(self):
        s = self.s
        del s.metadata.Acquisition_instrument
        with pytest.raises(ValueError):
            s.create_model()

    def test_auto_add_edges_true(self):
        m = self.s.create_model(auto_add_edges=True)
        cnames = [component.name for component in m]
        assert "B_K" in cnames and "C_K" in cnames

    def test_gos(self):
        m = self.s.create_model(auto_add_edges=True, GOS="hydrogenic")
        assert m["B_K"].GOS._name == "hydrogenic"

    def test_auto_add_background_true(self):
        m = self.s.create_model(auto_background=True)
        from hyperspy.components1d import PowerLaw
        is_pl_instance = [isinstance(c, PowerLaw) for c in m]
        assert True in is_pl_instance

    def test_auto_add_edges_false(self):
        m = self.s.create_model(auto_background=False)
        from hyperspy.components1d import PowerLaw
        is_pl_instance = [isinstance(c, PowerLaw) for c in m]
        assert not True in is_pl_instance

    def test_auto_add_edges_false_names(self):
        m = self.s.create_model(auto_add_edges=False)
        cnames = [component.name for component in m]
        assert not "B_K" in cnames or "C_K" in cnames

    def test_low_loss(self):
        ll = self.s.deepcopy()
        ll.axes_manager[-1].offset = -20
        m = self.s.create_model(ll=ll)
        assert m.low_loss is ll
        assert m.convolved

    def test_low_loss_bad_shape(self):
        ll = self.s.deepcopy()
        ll.axes_manager[-1].offset = -20
        ll.axes_manager.navigation_shape = (123,)
        with pytest.raises(ValueError):
            _ = self.s.create_model(ll=ll)


@lazifyTestClass
class TestEELSModel:

    def setup_method(self, method):
        s = hs.signals.EELSSpectrum(np.zeros(200))
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        s.add_elements(("B", "C"))
        self.m = s.create_model()

    def test_suspend_auto_fsw(self):
        m = self.m
        m["B_K"].fine_structure_width = 140.
        m.suspend_auto_fine_structure_width()
        m.enable_fine_structure()
        m.resolve_fine_structure()
        assert 140 == m["B_K"].fine_structure_width

    def test_resume_fsw(self):
        m = self.m
        m["B_K"].fine_structure_width = 140.
        m.suspend_auto_fine_structure_width()
        m.resume_auto_fine_structure_width()
        window = (m["C_K"].onset_energy.value -
                  m["B_K"].onset_energy.value - m._preedge_safe_window_width)
        m.enable_fine_structure()
        m.resolve_fine_structure()
        assert window == m["B_K"].fine_structure_width

    def test_get_first_ionization_edge_energy_C_B(self):
        assert (self.m._get_first_ionization_edge_energy() ==
                self.m["B_K"].onset_energy.value)

    def test_get_first_ionization_edge_energy_C(self):
        self.m["B_K"].active = False
        assert (self.m._get_first_ionization_edge_energy() ==
                self.m["C_K"].onset_energy.value)

    def test_get_first_ionization_edge_energy_None(self):
        self.m["B_K"].active = False
        self.m["C_K"].active = False
        assert self.m._get_first_ionization_edge_energy() is None

    def test_two_area_powerlaw_estimation_BC(self):
        self.m.signal.data = 2. * self.m.axis.axis ** (-3)  # A= 2, r=3
        self.m.signal.metadata.Signal.binned = False
        self.m.two_area_background_estimation()
        np.testing.assert_allclose(
            self.m._background_components[0].A.value,
            2.1451237089380295)
        np.testing.assert_allclose(
            self.m._background_components[0].r.value,
            3.0118980767392736)

    def test_two_area_powerlaw_estimation_C(self):
        self.m["B_K"].active = False
        self.m.signal.data = 2. * self.m.axis.axis ** (-3)  # A= 2, r=3
        self.m.signal.metadata.Signal.binned = False
        self.m.two_area_background_estimation()
        np.testing.assert_allclose(
            self.m._background_components[0].A.value,
            2.3978438900878087)
        np.testing.assert_allclose(
            self.m._background_components[0].r.value,
            3.031884021065014)

    def test_two_area_powerlaw_estimation_no_edge(self):
        self.m["B_K"].active = False
        self.m["C_K"].active = False
        self.m.signal.data = 2. * self.m.axis.axis ** (-3)  # A= 2, r=3
        self.m.signal.metadata.Signal.binned = False
        self.m.two_area_background_estimation()
        np.testing.assert_allclose(
            self.m._background_components[0].A.value,
            2.6598803469440986)
        np.testing.assert_allclose(
            self.m._background_components[0].r.value,
            3.0494030409062058)

    def test_get_start_energy_none(self):
        assert (self.m._get_start_energy() ==
                150)

    def test_get_start_energy_above(self):
        assert (self.m._get_start_energy(170) ==
                170)

    def test_get_start_energy_below(self):
        assert (self.m._get_start_energy(100) ==
                150)


@lazifyTestClass
class TestFitBackground:

    def setup_method(self, method):
        s = hs.signals.EELSSpectrum(np.ones(200))
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        CE = hs.material.elements.C.Atomic_properties.Binding_energies.K.onset_energy_eV
        BE = hs.material.elements.B.Atomic_properties.Binding_energies.K.onset_energy_eV
        s.isig[BE:] += 1
        s.isig[CE:] += 1
        s.add_elements(("Be", "B", "C"))
        self.m = s.create_model(auto_background=False)
        self.m.append(hs.model.components1D.Offset())

    def test_fit_background_B_C(self):
        self.m.fit_background()
        np.testing.assert_allclose(self.m["Offset"].offset.value,
                        1)
        assert self.m["B_K"].active
        assert self.m["C_K"].active

    def test_fit_background_C(self):
        self.m["B_K"].active = False
        self.m.fit_background()
        np.testing.assert_allclose(self.m["Offset"].offset.value,
                        1.7142857)
        assert not self.m["B_K"].active
        assert self.m["C_K"].active

    def test_fit_background_no_edge(self):
        self.m["B_K"].active = False
        self.m["C_K"].active = False
        self.m.fit_background()
        np.testing.assert_allclose(self.m["Offset"].offset.value,
                        2.14)
        assert not self.m["B_K"].active
        assert not self.m["C_K"].active
